import pkgutil
import sys
from importlib import import_module, reload

from django.apps import apps
from django.conf import settings
from django.db.migrations.graph import MigrationGraph
from django.db.migrations.recorder import MigrationRecorder

from .exceptions import (
    AmbiguityError, BadMigrationError, InconsistentMigrationHistory,
    NodeNotFoundError,
)

MIGRATIONS_MODULE_NAME = 'migrations'


class MigrationLoader:
    """
    Load migration files from disk and their status from the database.

    Migration files are expected to live in the "migrations" directory of
    an app. Their names are entirely unimportant from a code perspective,
    but will probably follow the 1234_name.py convention.

    On initialization, this class will scan those directories, and open and
    read the Python files, looking for a class called Migration, which should
    inherit from django.db.migrations.Migration. See
    django.db.migrations.migration for what that looks like.

    Some migrations will be marked as "replacing" another set of migrations.
    These are loaded into a separate set of migrations away from the main ones.
    If all the migrations they replace are either unapplied or missing from
    disk, then they are injected into the main set, replacing the named migrations.
    Any dependency pointers to the replaced migrations are re-pointed to the
    new migration.

    This does mean that this class MUST also talk to the database as well as
    to disk, but this is probably fine. We're already not just operating
    in memory.
    """

    def __init__(self, connection, load=True, ignore_no_migrations=False):
        self.connection = connection
        self.disk_migrations = None
        self.applied_migrations = None
        self.ignore_no_migrations = ignore_no_migrations
        if load:
            self.build_graph()

    @classmethod
    def migrations_module(cls, app_label):
        """
        Return the path to the migrations module for the specified app_label
        and a boolean indicating if the module is specified in
        settings.MIGRATION_MODULE.
        """
        if app_label in settings.MIGRATION_MODULES:
            return settings.MIGRATION_MODULES[app_label], True
        else:
            app_package_name = apps.get_app_config(app_label).name
            return '%s.%s' % (app_package_name, MIGRATIONS_MODULE_NAME), False

    def load_disk(self):
        """Load the migrations from all INSTALLED_APPS from disk."""
        self.disk_migrations = {}
        self.unmigrated_apps = set()
        self.migrated_apps = set()
        for app_config in apps.get_app_configs():
            # Get the migrations module directory
            module_name, explicit = self.migrations_module(app_config.label)
            if module_name is None:
                self.unmigrated_apps.add(app_config.label)
                continue
            was_loaded = module_name in sys.modules
            try:
                module = import_module(module_name)
            except ImportError as e:
                # I hate doing this, but I don't want to squash other import errors.
                # Might be better to try a directory check directly.
                if ((explicit and self.ignore_no_migrations) or (
                        not explicit and "No module named" in str(e) and MIGRATIONS_MODULE_NAME in str(e))):
                    self.unmigrated_apps.add(app_config.label)
                    continue
                raise
            else:
                # Empty directories are namespaces.
                # getattr() needed on PY36 and older (replace w/attribute access).
                if getattr(module, '__file__', None) is None:
                    self.unmigrated_apps.add(app_config.label)
                    continue
                # Module is not a package (e.g. migrations.py).
                if not hasattr(module, '__path__'):
                    self.unmigrated_apps.add(app_config.label)
                    continue
                # Force a reload if it's already loaded (tests need this)
                if was_loaded:
                    reload(module)
            self.migrated_apps.add(app_config.label)
            migration_names = {
                name for _, name, is_pkg in pkgutil.iter_modules(module.__path__)
                if not is_pkg and name[0] not in '_~'
            }
            # Load migrations
            for migration_name in migration_names:
                migration_path = '%s.%s' % (module_name, migration_name)
                try:
                    migration_module = import_module(migration_path)
                except ImportError as e:
                    if 'bad magic number' in str(e):
                        raise ImportError(
                            "Couldn't import %r as it appears to be a stale "
                            ".pyc file." % migration_path
                        ) from e
                    else:
                        raise
                if not hasattr(migration_module, "Migration"):
                    raise BadMigrationError(
                        "Migration %s in app %s has no Migration class" % (migration_name, app_config.label)
                    )
                self.disk_migrations[app_config.label, migration_name] = migration_module.Migration(
                    migration_name,
                    app_config.label,
                )

    def get_migration(self, app_label, name_prefix):
        """Return the named migration or raise NodeNotFoundError."""
        return self.graph.nodes[app_label, name_prefix]

    def get_migration_by_prefix(self, app_label, name_prefix):
        """
        Return the migration(s) which match the given app label and name_prefix.
        """
        # Do the search
        results = []
        for migration_app_label, migration_name in self.disk_migrations:
            if migration_app_label == app_label and migration_name.startswith(name_prefix):
                results.append((migration_app_label, migration_name))
        if len(results) > 1:
            raise AmbiguityError(
                "There is more than one migration for '%s' with the prefix '%s'" % (app_label, name_prefix)
            )
        elif not results:
            raise KeyError("There no migrations for '%s' with the prefix '%s'" % (app_label, name_prefix))
        else:
            return self.disk_migrations[results[0]]

    def check_key(self, key, current_app):
        if (key[1] != "__first__" and key[1] != "__latest__") or key in self.graph:
            return key
        # Special-case __first__, which means "the first migration" for
        # migrated apps, and is ignored for unmigrated apps. It allows
        # makemigrations to declare dependencies on apps before they even have
        # migrations.
        if key[0] == current_app:
            # Ignore __first__ references to the same app (#22325)
            return
        if key[0] in self.unmigrated_apps:
            # This app isn't migrated, but something depends on it.
            # The models will get auto-added into the state, though
            # so we're fine.
            return
        if key[0] in self.migrated_apps:
            try:
                if key[1] == "__first__":
                    return self.graph.root_nodes(key[0])[0]
                else:  # "__latest__"
                    return self.graph.leaf_nodes(key[0])[0]
            except IndexError:
                if self.ignore_no_migrations:
                    return None
                else:
                    raise ValueError("Dependency on app with no migrations: %s" % key[0])
        raise ValueError("Dependency on unknown app: %s" % key[0])

    def add_internal_dependencies(self, key, migration):
        """
        Internal dependencies need to be added first to ensure `__first__`
        dependencies find the correct root node.
        """
        for parent in migration.dependencies:
            # Ignore __first__ references to the same app.
            if parent[0] == key[0] and parent[1] != '__first__':
                self.graph.add_dependency(migration, key, parent, skip_validation=True)

    def add_external_dependencies(self, key, migration):
        for parent in migration.dependencies:
            # Skip internal dependencies
            if key[0] == parent[0]:
                continue
            parent = self.check_key(parent, key[0])
            if parent is not None:
                self.graph.add_dependency(migration, key, parent, skip_validation=True)
        for child in migration.run_before:
            child = self.check_key(child, key[0])
            if child is not None:
                self.graph.add_dependency(migration, child, key, skip_validation=True)

    def build_graph(self):
        """
        Build a migration dependency graph using both the disk and database.
        You'll need to rebuild the graph if you apply migrations. This isn't
        usually a problem as generally migration stuff runs in a one-shot process.
        """
        # Load disk data
        self.load_disk()
        # Load database data
        if self.connection is None:
            self.applied_migrations = set()
        else:
            recorder = MigrationRecorder(self.connection)
            self.applied_migrations = recorder.applied_migrations()
        # To start, populate the migration graph with nodes for ALL migrations
        # and their dependencies. Also make note of replacing migrations at this step.
        self.graph = MigrationGraph()
        self.replacements = {}
        for key, migration in self.disk_migrations.items():
            self.graph.add_node(key, migration)
            # Replacing migrations.
            if migration.replaces:
                self.replacements[key] = migration
        for key, migration in self.disk_migrations.items():
            # Internal (same app) dependencies.
            self.add_internal_dependencies(key, migration)
        # Add external dependencies now that the internal ones have been resolved.
        for key, migration in self.disk_migrations.items():
            self.add_external_dependencies(key, migration)
        # Carry out replacements where possible.
        for key, migration in self.replacements.items():
            # Get applied status of each of this migration's replacement targets.
            applied_statuses = [(target in self.applied_migrations) for target in migration.replaces]
            # Ensure the replacing migration is only marked as applied if all of
            # its replacement targets are.
            if all(applied_statuses):
                self.applied_migrations.add(key)
            else:
                self.applied_migrations.discard(key)
            # A replacing migration can be used if either all or none of its
            # replacement targets have been applied.
            if all(applied_statuses) or (not any(applied_statuses)):
                self.graph.remove_replaced_nodes(key, migration.replaces)
            else:
                # This replacing migration cannot be used because it is partially applied.
                # Remove it from the graph and remap dependencies to it (#25945).
                self.graph.remove_replacement_node(key, migration.replaces)
        # Ensure the graph is consistent.
        try:
            self.graph.validate_consistency()
        except NodeNotFoundError as exc:
            # Check if the missing node could have been replaced by any squash
            # migration but wasn't because the squash migration was partially
            # applied before. In that case raise a more understandable exception
            # (#23556).
            # Get reverse replacements.
            reverse_replacements = {}
            for key, migration in self.replacements.items():
                for replaced in migration.replaces:
                    reverse_replacements.setdefault(replaced, set()).add(key)
            # Try to reraise exception with more detail.
            if exc.node in reverse_replacements:
                candidates = reverse_replacements.get(exc.node, set())
                is_replaced = any(candidate in self.graph.nodes for candidate in candidates)
                if not is_replaced:
                    tries = ', '.join('%s.%s' % c for c in candidates)
                    raise NodeNotFoundError(
                        "Migration {0} depends on nonexistent node ('{1}', '{2}'). "
                        "Django tried to replace migration {1}.{2} with any of [{3}] "
                        "but wasn't able to because some of the replaced migrations "
                        "are already applied.".format(
                            exc.origin, exc.node[0], exc.node[1], tries
                        ),
                        exc.node
                    ) from exc
            raise exc
        self.graph.ensure_not_cyclic()

    def check_consistent_history(self, connection):
        """
        Raise InconsistentMigrationHistory if any applied migrations have
        unapplied dependencies.
        """
        recorder = MigrationRecorder(connection)
        applied = recorder.applied_migrations()
        for migration in applied:
            # If the migration is unknown, skip it.
            if migration not in self.graph.nodes:
                continue
            for parent in self.graph.node_map[migration].parents:
                if parent not in applied:
                    # Skip unapplied squashed migrations that have all of their
                    # `replaces` applied.
                    if parent in self.replacements:
                        if all(m in applied for m in self.replacements[parent].replaces):
                            continue
                    raise InconsistentMigrationHistory(
                        "Migration {}.{} is applied before its dependency "
                        "{}.{} on database '{}'.".format(
                            migration[0], migration[1], parent[0], parent[1],
                            connection.alias,
                        )
                    )

    def detect_conflicts(self):
        """
        Look through the loaded graph and detect any conflicts - apps
        with more than one leaf migration. Return a dict of the app labels
        that conflict with the migration names that conflict.
        """
        seen_apps = {}
        conflicting_apps = set()
        for app_label, migration_name in self.graph.leaf_nodes():
            if app_label in seen_apps:
                conflicting_apps.add(app_label)
            seen_apps.setdefault(app_label, set()).add(migration_name)
        return {app_label: seen_apps[app_label] for app_label in conflicting_apps}

    def project_state(self, nodes=None, at_end=True):
        """
        Return a ProjectState object representing the most recent state
        that the loaded migrations represent.

        See graph.make_state() for the meaning of "nodes" and "at_end".
        """
        return self.graph.make_state(nodes=nodes, at_end=at_end, real_apps=list(self.unmigrated_apps))



from .. import Provider as InternetProvider


class Provider(InternetProvider):
    user_name_formats = (
        '{{last_name_female}}.{{first_name_female}}',
        '{{last_name_female}}.{{first_name_female}}',
        '{{last_name_male}}.{{first_name_male}}',
        '{{last_name_male}}.{{first_name_male}}',
        '{{first_name_female}}.{{last_name_female}}',
        '{{first_name_male}}.{{last_name_male}}',
        '{{first_name}}##',
        '?{{last_name}}',
        '?{{last_name}}',
        '?{{last_name}}',
    )

    email_formats = ('{{user_name}}@{{free_email_domain}}', )

    free_email_domains = (
        'zoznam.sk',
        'gmail.com',
        'centrum.sk',
        'post.sk',
        'chello.sk',
        'pobox.sk',
        'szm.sk',
        'atlas.sk',
        'azet.sk',
        'inmail.sk',
    )

    tlds = ('sk', 'com')



from django import VERSION as DJANGO_VERSION
from django.conf import settings
from django.db import connection, transaction, utils

try:
    from django.db.models.signals import post_migrate, post_syncdb
except ImportError:
    from django.db.models.signals import post_migrate

VIEW_SQL = "CREATE OR REPLACE VIEW %(schema)s.%(model_table)s as select * from %(version_model)s inner join %(base_model)s on %(base_model)s.id = %(version_model)s.object_id"
DROP_SQL = "DROP VIEW IF EXISTS %(schema)s.%(model_table)s;"
EXISTS = "SELECT exists(select schema_name FROM information_schema.schemata WHERE schema_name = %s)"
TRIGGER = "CREATE TRIGGER %(trigger_name)s INSTEAD OF UPDATE OR DELETE ON %(schema)s.%(model_table)s FOR EACH ROW EXECUTE PROCEDURE %(function_name)s();"


def trigger_function(base_model, version_model, args):
    qn = connection.ops.quote_name
    base_cols = [
        f.column for f in base_model._meta.local_fields if f.column and f.column != "id"
    ]
    version_cols = [
        f.column
        for f in version_model._meta.local_fields
        if f.column and f.column != "vid"
    ]

    local_args = dict(args)
    local_args.update(
        {
            "base_cols": "id=OLD.id,"
            + ", ".join([f"{qn(x)}=NEW.{qn(x)}" for x in base_cols]),
            "version_cols": "vid=OLD.vid,"
            + ", ".join([f"{qn(x)}=NEW.{qn(x)}" for x in version_cols]),
        }
    )

    return (
        """
    CREATE OR REPLACE FUNCTION %(function_name)s()
    RETURNS TRIGGER
    LANGUAGE plpgsql
    AS $function$
        BEGIN
            IF TG_OP = 'UPDATE' THEN
                UPDATE %(base_model)s SET %(base_cols)s WHERE id=OLD.id;
                UPDATE %(version_model)s SET %(version_cols)s WHERE object_id=OLD.id and vid=OLD.vid;
                RETURN NEW;
            ELSIF TG_OP = 'DELETE' THEN
                DELETE FROM %(base_model)s WHERE id=OLD.id;
                DELETE FROM %(version_model)s WHERE object_id=OLD.id;
                RETURN NULL;
            END IF;
        END;
    $function$;
    """
        % local_args
    )


def do_updates(m):
    if getattr(m._meta, "_is_view", None):
        qn = connection.ops.quote_name
        args = {
            "base_model": qn(m._meta._base_model._meta.db_table),
            "version_model": qn(m._meta._version_model._meta.db_table),
            "model_table": qn(m._meta.db_table),
            "function_name": qn(m._meta.db_table + "_func"),
            "trigger_name": qn(m._meta.db_table + "_trigger"),
            "schema": qn("public"),
            "state": qn("state"),
        }

        base_sql = VIEW_SQL % args
        cursor = connection.cursor()

        cursor.execute(DROP_SQL % args)
        cursor.execute(base_sql)

        trigger_sql = trigger_function(
            m._meta._base_model, m._meta._version_model, args
        )
        cursor.execute(trigger_sql)
        cursor.execute(TRIGGER % args)
        for schema in m._meta._version_model.UNIQUE_STATES:
            qn_schema = qn(schema)

            # Make sure schema exists
            cursor.execute(
                "SELECT exists(select schema_name FROM information_schema.schemata WHERE schema_name = %s)",
                (schema,),
            )
            if not cursor.fetchone()[0]:
                cursor.execute(f"CREATE SCHEMA {qn_schema}")

            args["schema"] = qn_schema

            cursor.execute(DROP_SQL % args)
            where = "WHERE %(version_model)s.%(state)s = %%s" % args

            base_sql = VIEW_SQL % args
            sql = " ".join([base_sql, where])
            cursor.execute(sql, (schema,))
            cursor.execute(TRIGGER % args)

        if DJANGO_VERSION < (1, 6):
            transaction.commit_unless_managed()


def update_schema(sender=None, **kwargs):
    if sender:
        for m in sender.get_models():
            if getattr(m._meta, "_view_model", None):
                do_updates(m._meta._view_model)


def update_syncdb_schema(app, created_models, verbosity, **kwargs):
    for m in created_models:
        do_updates(m)


if DJANGO_VERSION < (1, 7):
    post_syncdb.connect(update_syncdb_schema, dispatch_uid="update_syncdb_schema")
else:
    post_migrate.connect(update_schema, dispatch_uid="update_schema")



# Generated by Django 2.2.16 on 2020-09-17 11:20
import taggit.managers
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('assets', '0002_auto_20200330_2224'),
    ]

    operations = [
        migrations.AlterField(
            model_name='asset',
            name='tags',
            field=taggit.managers.TaggableManager(help_text='Search tags are used to prefilter assets when including an asset on your object', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
    ]



from functools import update_wrapper

from django import forms
from django.contrib.admin.forms import AdminAuthenticationForm
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.paginator import Paginator
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.translation import ugettext as _
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect

from . import models
from .forms import BaseFilterForm

try:
    try:
        from ..versioning import manager
        from ..versioning.models import BaseVersionedModel
    except ValueError:
        from versioning import manager
        from versioning.models import BaseVersionedModel
except ImportError:
    manager = None

LOGIN_FORM_KEY = "this_is_the_login_form"


class AlreadyRegistered(Exception):
    pass


class NotRegistered(Exception):
    pass


class AdminSite:
    """
    Heavily copied from django's admin site
    but instead of registering models
    we register CMSBundle objects.

    :param login_form: A Form class to use for the login form. \
    Defaults to django's AdminAuthenticationForm.
    :param login_template: Template for login page. Defaults to \
    'cms/login.html'.
    :param logout_template: Template for logout page. Defaults to \
    'cms/logged_out.html'.
    :param password_change_template: Template for password change page. \
    Defaults to 'cms/password_change_form.html'.
    :param password_change_done_template: Template displayed after \
    password is changed. Defaults to 'cms/password_change_done.html'.
    :param dashboard_template: Dashboard template. Defaults to \
    cms/dashboard.html.
    :param dashboard_home_url: Dashboard Homepage URL. Defaults to /admin/.
    """

    login_form = None
    login_template = None
    logout_template = None
    password_change_template = None
    password_change_done_template = None
    dashboard_template = None

    def __init__(self, name="default"):
        self._registry = {}
        self._model_registry = {}
        self._titles = {}
        self._order = {}
        self.name = name

    def register_model(self, model, bundle):
        """
        Registers a bundle as the main bundle for a
        model. Used when we need to lookup urls by
        a model.
        """
        if model in self._model_registry:
            raise AlreadyRegistered(f"The model {model} is already registered")

        if bundle.url_params:
            raise Exception(
                "A primary model bundle cannot have dynamic \
                            url_parameters"
            )

        self._model_registry[model] = bundle

    def get_bundle_for_model(self, model):
        """
        Returns the main bundle for the given
        model
        """
        return self._model_registry.get(model)

    def unregister_model(self, model):
        """
        Unregisters the given model.
        """
        if model not in self._model_registry:
            raise NotRegistered(f"The model {model} is not registered")

        del self._model_registry[model]

    def register(self, slug, bundle, order=1, title=None):
        """
        Registers the bundle for a certain slug.

        If a slug is already registered, this will raise AlreadyRegistered.

        :param slug: The slug to register.
        :param bundle: The bundle instance being registered.
        :param order: An integer that controls where this bundle's \
        dashboard links appear in relation to others.
        """

        if slug in self._registry:
            raise AlreadyRegistered(f"The url {slug} is already registered")

        # Instantiate the admin class to save in the registry.
        self._registry[slug] = bundle
        self._order[slug] = order
        if title:
            self._titles[slug] = title
        bundle.set_admin_site(self)

    def unregister(self, slug):
        """
        Unregisters the given url.

        If a slug isn't already registered, this will raise NotRegistered.
        """

        if slug not in self._registry:
            raise NotRegistered(f"The slug {slug} is not registered")
        bundle = self._registry[slug]
        if bundle._meta.model and bundle._meta.primary_model_bundle:
            self.unregister_model(bundle._meta.model)

        del self._registry[slug]
        del self._order[slug]

    def has_permission(self, request):
        """
        Returns True if the given HttpRequest has permission to view
        *at least one* page in the admin site.

        Currently checks that the user is active and a staff member.
        """
        return request.user.is_active and request.user.is_staff

    def admin_view(self, view, cacheable=False):
        def inner(request, *args, **kwargs):
            if manager:
                manager.activate(BaseVersionedModel.DRAFT)
            if not self.has_permission(request):
                if request.path == reverse("admin:cms_logout"):
                    index_path = reverse("admin:cms_index")
                    return HttpResponseRedirect(index_path)
                return self.login(request)
            return view(request, *args, **kwargs)

        if not cacheable:
            inner = never_cache(inner)
        # We add csrf_protect here so this function can be used as a utility
        # function for any view, without having to repeat 'csrf_protect'.
        if not getattr(view, "csrf_exempt", False):
            inner = csrf_protect(inner)
        return update_wrapper(inner, view)

    def get_urls(self):
        from django.conf.urls import include, url

        def wrap(view, cacheable=False):
            def wrapper(*args, **kwargs):
                return self.admin_view(view, cacheable)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        # Admin-site-wide views.
        urlpatterns = [
            url(r"^$", wrap(self.index), name="cms_index"),
            url(r"^logout/$", wrap(self.logout), name="cms_logout"),
            url(
                r"^password_change/$",
                wrap(self.password_change, cacheable=True),
                name="cms_password_change",
            ),
            url(
                r"^password_change/done/$",
                wrap(self.password_change_done, cacheable=True),
                name="cms_password_change_done",
            ),
        ]

        # Add in each model's views.
        for base, bundle in self._registry.items():
            urlpatterns += [url(rf"^{base}/", include(bundle.get_urls()))]
        return urlpatterns

    @property
    def urls(self):
        return self.get_urls(), "admin", self.name

    def password_change(self, request):
        """
        Handles the "change password" task -- both form display and validation.

        Uses the default auth views.
        """
        from django.contrib.auth.views import PasswordChangeView

        url = reverse("admin:cms_password_change_done")
        defaults = {
            "success_url": url,
            "template_name": "cms/password_change_form.html",
        }
        if self.password_change_template is not None:
            defaults["template_name"] = self.password_change_template
        return PasswordChangeView.as_view(**defaults)(request)

    def password_change_done(self, request, extra_context=None):
        """
        Displays the "success" page after a password change.
        """
        from django.contrib.auth.views import PasswordChangeDoneView

        defaults = {
            "extra_context": extra_context or {},
            "template_name": "cms/password_change_done.html",
        }
        if self.password_change_done_template is not None:
            defaults["template_name"] = self.password_change_done_template
        return PasswordChangeDoneView.as_view(**defaults)(request)

    @never_cache
    def logout(self, request, extra_context=None):
        """
        Logs out the user for the given HttpRequest.

        This should *not* assume the user is already logged in.
        """
        from django.contrib.auth.views import LogoutView

        defaults = {
            "extra_context": extra_context or {},
            "template_name": "cms/logged_out.html",
        }
        if self.logout_template is not None:
            defaults["template_name"] = self.logout_template
        return LogoutView.as_view(**defaults)(request)

    @never_cache
    def login(self, request, extra_context=None):
        """
        Displays the login form for the given HttpRequest.
        """
        from django.contrib.auth.views import LoginView

        context = {
            "title": _("Log in"),
            "app_path": request.get_full_path(),
            REDIRECT_FIELD_NAME: request.get_full_path(),
        }
        context.update(extra_context or {})
        defaults = {
            "extra_context": context,
            "authentication_form": self.login_form or AdminAuthenticationForm,
            "template_name": self.login_template or "cms/login.html",
        }
        return LoginView.as_view(**defaults)(request)

    def get_dashboard_urls(self, request):
        nav = []
        for k in sorted(self._order, key=self._order.get):
            v = self._registry[k]
            urls = v.get_dashboard_urls(request)
            if urls:
                title = self._titles.get(k, v.get_title())
                nav.append((title, urls, v.name))
        return nav

    def get_dashboard_blocks(self, request):
        blocks = []
        for k in sorted(self._order, key=self._order.get):
            v = self._registry[k]
            block = v.get_dashboard_block(request)
            if block:
                title = self._titles.get(k, v.get_title())
                blocks.append((title, block))
        return blocks

    def _get_allowed_sections(self, dashboard):
        """
        Get the sections to display based on dashboard
        """

        allowed_titles = [x[0] for x in dashboard]
        allowed_sections = [x[2] for x in dashboard]
        return tuple(allowed_sections), tuple(allowed_titles)

    @never_cache
    def index(self, request, extra_context=None):
        """
        Displays the dashboard. Includes the main
        navigation that the user has permission for as well
        as the cms log for those sections. The log list can
        be filtered by those same sections
        and is paginated.
        """

        dashboard = self.get_dashboard_urls(request)
        dash_blocks = self.get_dashboard_blocks(request)

        sections, titles = self._get_allowed_sections(dashboard)
        choices = list(zip(sections, titles))
        choices.sort(key=lambda tup: tup[1])
        choices.insert(0, ("", "All"))

        class SectionFilterForm(BaseFilterForm):
            section = forms.ChoiceField(required=False, choices=choices)

        form = SectionFilterForm(request.GET)
        filter_kwargs = form.get_filter_kwargs()

        if not filter_kwargs and not request.user.is_superuser:
            filter_kwargs["section__in"] = sections
        cms_logs = models.CMSLog.objects.filter(**filter_kwargs).order_by("-when")

        template = self.dashboard_template or "cms/dashboard.html"

        paginator = Paginator(cms_logs[: 20 * 100], 20, allow_empty_first_page=True)
        page_number = request.GET.get("page") or 1
        try:
            page_number = int(page_number)
        except ValueError:
            page_number = 1

        page = paginator.page(page_number)

        return TemplateResponse(
            request,
            [template],
            {
                "dashboard": dashboard,
                "blocks": dash_blocks,
                "page": page,
                "bundle": list(self._registry.values())[0],
                "form": form,
            },
        )


# This global object represents the default admin site, for the common case.
# You can instantiate AdminSite in your own code to create a custom admin site.
site = AdminSite()



from faker import Faker
import pytest

from scarlet.cache.groups import CacheGroup
from scarlet.cache.router import router

from .models import DummyModel


fake = Faker()
CACHE_KEY = fake.pystr()
CACHE = router.get_cache()
EXTRA = "dummy"


def teardown_function():
    CACHE.clear()


@pytest.fixture
def cache_group():
    return CacheGroup(CACHE_KEY)


def test_init():
    cg = CacheGroup(key=CACHE_KEY)
    assert cg.key == CACHE_KEY
    assert cg.version_expiry

    cg = CacheGroup(key=CACHE_KEY, version_expiry=100)
    assert cg.key == CACHE_KEY
    assert cg.version_expiry == 100


def test_init_no_key():
    with pytest.raises(AssertionError):
        CacheGroup()

    with pytest.raises(AssertionError):
        CacheGroup(key=None)


def test_register(cache_group):
    cache_group.register(DummyModel)
    assert cache_group.models == [DummyModel]


def test_register_models(cache_group):
    fake_str = fake.pystr()
    fake_str2 = fake.pystr()
    cache_group.register_models(fake_str, fake_str2)
    assert cache_group.models == [fake_str, fake_str2]


def test_double_register_models(cache_group):
    fake_str = fake.pystr()
    fake_str2 = fake.pystr()
    with pytest.raises(Exception):
        cache_group.register_models(fake_str, fake_str2, fake_str)


def test_double_register(cache_group):
    cache_group.register(DummyModel)
    with pytest.raises(Exception):
        cache_group.register(DummyModel)


def test_get_version(cache_group):
    cache_group.register(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.0"
    assert cache_group.get_version("") == f"{CACHE_KEY}.0"
    assert cache_group.get_version("1") == f"{CACHE_KEY}.0.1.0"
    assert cache_group.get_version("5") == f"{CACHE_KEY}.0.5.0"
    assert cache_group.get_version("test") == f"{CACHE_KEY}.0.test.0"
    assert (
        cache_group.get_version("test_underscore") == f"{CACHE_KEY}.0.test_underscore.0"
    )


def test_invalidate_cache(cache_group):
    cache_group.register(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.0"

    cache_group.invalidate_cache(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.1"


def test_invalidate_cache_nonexistent_key(cache_group):
    cache_group.register(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.0"

    cache_group.invalidate_cache("nonexistent_key")
    assert cache_group.get_version() == f"{CACHE_KEY}.0"


def test_invalidate_cache_reset_when_val_over_threshold(cache_group):
    cache_group.register(DummyModel)

    CACHE.set(CACHE_KEY, float("inf"))
    cache_group.invalidate_cache(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.0"


def test_invalidate_cache_extra(cache_group):
    cache_group.register(DummyModel, values=[EXTRA])
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.0.{EXTRA}.0"

    cache_group.invalidate_cache(DummyModel)
    assert cache_group.get_version() == f"{CACHE_KEY}.0"
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.0.{EXTRA}.1"


def test_invalidate_cache_instance_values(cache_group):
    cache_group.register(DummyModel, instance_values=["pk"])
    assert cache_group.get_version("2") == f"{CACHE_KEY}.0.2.0"

    cache_group.invalidate_cache(DummyModel, instance=DummyModel(pk=2))
    assert cache_group.get_version() == f"{CACHE_KEY}.0"
    assert cache_group.get_version("2") == f"{CACHE_KEY}.0.2.1"


def test_invalidate_cache_extra_and_instance_values(cache_group):
    cache_group.register(DummyModel, values=[EXTRA], instance_values=["pk"])
    assert cache_group.get_version() == f"{CACHE_KEY}.0"
    assert cache_group.get_version("2") == f"{CACHE_KEY}.0.2.0"
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.0.{EXTRA}.0"

    cache_group.invalidate_cache(DummyModel, instance=DummyModel(pk=2), extra=[EXTRA])
    assert cache_group.get_version() == f"{CACHE_KEY}.0"
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.0.{EXTRA}.2"
    assert cache_group.get_version("2") == f"{CACHE_KEY}.0.2.1"


def test_invalidate_cache_force_all(cache_group):
    cache_group.register(DummyModel, values=[EXTRA], instance_values=["pk"])
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.0.{EXTRA}.0"
    assert cache_group.get_version("2") == f"{CACHE_KEY}.0.2.0"

    cache_group.invalidate_cache(DummyModel, force_all=True)
    assert cache_group.get_version() == f"{CACHE_KEY}.1"
    assert cache_group.get_version("2") == f"{CACHE_KEY}.1.2.0"
    assert cache_group.get_version(EXTRA) == f"{CACHE_KEY}.1.{EXTRA}.0"



import pytest

from faker import Faker
from faker.config import DEFAULT_LOCALE

DEFAULT_SEED = 0


@pytest.fixture(scope='session', autouse=True)
def _session_faker(request):
    """Fixture that stores the session level ``Faker`` instance.

    This fixture is internal and is only meant for use within the project.
    Third parties should instead use the ``faker`` fixture for their tests.
    """
    if 'faker_session_locale' in request.fixturenames:
        locale = request.getfixturevalue('faker_session_locale')
    else:
        locale = [DEFAULT_LOCALE]
    return Faker(locale=locale)


@pytest.fixture()
def faker(request):
    """Fixture that returns a seeded and suitable ``Faker`` instance."""
    if 'faker_locale' in request.fixturenames:
        locale = request.getfixturevalue('faker_locale')
        fake = Faker(locale=locale)
    else:
        fake = request.getfixturevalue('_session_faker')

    seed = DEFAULT_SEED
    if 'faker_seed' in request.fixturenames:
        seed = request.getfixturevalue('faker_seed')
    fake.seed_instance(seed=seed)

    return fake



import importlib

from django.apps import apps as djapps
from django.core.exceptions import ImproperlyConfigured

from . import settings

default_app_config = "scarlet.assets.apps.AppConfig"


def get_asset_model():
    try:
        app_label, model_name = settings.ASSET_MODEL.rsplit(".", 1)
    except ValueError:
        raise ImproperlyConfigured(
            "ASSET_MODEL must be of the form 'app_label.model_name'"
        )

    return djapps.get_model(app_label, model_name)


def get_image_cropper():
    parts = settings.IMAGE_CROPPER.rsplit(".", 1)
    importpath = parts[0]
    attr = parts[-1]
    if importpath.startswith("."):
        mod = importlib.import_module(__name__ + importpath)
    else:
        mod = importlib.import_module(importpath)
    return getattr(mod, attr)



from ..es_ES import Provider as PersonProvider


class Provider(PersonProvider):
    """
    Adds popular Catalan names.
    https://www.idescat.cat/pub/?id=aec&n=946&lang=es&t=2018
    https://www.idescat.cat/pub/?id=aec&n=947&lang=es&t=2018
    """
    first_names_male = (
        'Adam',
        'Albert',
        'Aleix',
        'Álex',
        'Antonio',
        'Arnau',
        'Biel',
        'Bruno',
        'Carlos',
        'Daniel',
        'David',
        'Enzo',
        'Èric',
        'Francisco',
        'Hugo',
        'Jan',
        'Javier',
        'Joan',
        'Jordi',
        'Jorge',
        'Josep',
        'José',
        'José María',
        'Juan',
        'Leo',
        'Lucas',
        'Manuel',
        'Marc',
        'Martí',
        'Max',
        'Miguel',
        'Nil',
        'Pau',
        'Pedro',
        'Pol',
        'Ramón',
        'Xavier')

    first_names_female = (
        'Abril',
        'Aina',
        'Ana',
        'Anna',
        'Antonia',
        'Antònia',
        'Arlet',
        'Carla',
        'Carmen',
        'Chlóe',
        'Clàudia',
        'Cristina',
        'Dolores',
        'Emma',
        'Francisca',
        'Isabel',
        'Jana',
        'Josefa',
        'Júlia',
        'Laia',
        'Laura',
        'Lucia',
        'Marta',
        'Martina',
        'María',
        'María Del Carmen',
        'María Dolores',
        'María Teresa',
        'Mia',
        'Montserrat',
        'Noa',
        'Núria',
        'Ona',
        'Paula',
        'Rosa',
        'Sara',
        'Sofía',
        'Sílvia',
        'Valèria')

    first_names = first_names_male + first_names_female



import re
import string

from collections import OrderedDict

from faker.utils.distribution import choices_distribution, choices_distribution_unique

_re_hash = re.compile(r'#')
_re_perc = re.compile(r'%')
_re_excl = re.compile(r'!')
_re_at = re.compile(r'@')
_re_qm = re.compile(r'\?')
_re_cir = re.compile(r'\^')


class BaseProvider:

    __provider__ = 'base'
    __lang__ = None

    # Locales supported by Linux Mint from `/usr/share/i18n/SUPPORTED`
    language_locale_codes = {
        'aa': ('DJ', 'ER', 'ET'), 'af': ('ZA',), 'ak': ('GH',), 'am': ('ET',),
        'an': ('ES',), 'apn': ('IN',),
        'ar': ('AE', 'BH', 'DJ', 'DZ', 'EG', 'EH', 'ER', 'IL', 'IN',
               'IQ', 'JO', 'KM', 'KW', 'LB', 'LY', 'MA', 'MR', 'OM',
               'PS', 'QA', 'SA', 'SD', 'SO', 'SS', 'SY', 'TD', 'TN',
               'YE'),
        'as': ('IN',), 'ast': ('ES',), 'ayc': ('PE',), 'az': ('AZ', 'IN'),
        'be': ('BY',), 'bem': ('ZM',), 'ber': ('DZ', 'MA'), 'bg': ('BG',),
        'bhb': ('IN',), 'bho': ('IN',), 'bn': ('BD', 'IN'), 'bo': ('CN', 'IN'),
        'br': ('FR',), 'brx': ('IN',), 'bs': ('BA',), 'byn': ('ER',),
        'ca': ('AD', 'ES', 'FR', 'IT'), 'ce': ('RU',), 'ckb': ('IQ',),
        'cmn': ('TW',), 'crh': ('UA',), 'cs': ('CZ',), 'csb': ('PL',),
        'cv': ('RU',), 'cy': ('GB',), 'da': ('DK',),
        'de': ('AT', 'BE', 'CH', 'DE', 'LI', 'LU'), 'doi': ('IN',),
        'dv': ('MV',), 'dz': ('BT',), 'el': ('GR', 'CY'),
        'en': ('AG', 'AU', 'BW', 'CA', 'DK', 'GB', 'HK', 'IE', 'IN', 'NG',
               'NZ', 'PH', 'SG', 'US', 'ZA', 'ZM', 'ZW'),
        'eo': ('US',),
        'es': ('AR', 'BO', 'CL', 'CO', 'CR', 'CU', 'DO', 'EC', 'ES', 'GT',
               'HN', 'MX', 'NI', 'PA', 'PE', 'PR', 'PY', 'SV', 'US', 'UY', 'VE',
               ), 'et': ('EE',), 'eu': ('ES', 'FR'), 'fa': ('IR',),
        'ff': ('SN',), 'fi': ('FI',), 'fil': ('PH',), 'fo': ('FO',),
        'fr': ('CA', 'CH', 'FR', 'LU'), 'fur': ('IT',), 'fy': ('NL', 'DE'),
        'ga': ('IE',), 'gd': ('GB',), 'gez': ('ER', 'ET'), 'gl': ('ES',),
        'gu': ('IN',), 'gv': ('GB',), 'ha': ('NG',), 'hak': ('TW',),
        'he': ('IL',), 'hi': ('IN',), 'hne': ('IN',), 'hr': ('HR',),
        'hsb': ('DE',), 'ht': ('HT',), 'hu': ('HU',), 'hy': ('AM',),
        'ia': ('FR',), 'id': ('ID',), 'ig': ('NG',), 'ik': ('CA',),
        'is': ('IS',), 'it': ('CH', 'IT'), 'iu': ('CA',), 'iw': ('IL',),
        'ja': ('JP',), 'ka': ('GE',), 'kk': ('KZ',), 'kl': ('GL',),
        'km': ('KH',), 'kn': ('IN',), 'ko': ('KR',), 'kok': ('IN',),
        'ks': ('IN',), 'ku': ('TR',), 'kw': ('GB',), 'ky': ('KG',),
        'lb': ('LU',), 'lg': ('UG',), 'li': ('BE', 'NL'), 'lij': ('IT',),
        'ln': ('CD',), 'lo': ('LA',), 'lt': ('LT',), 'lv': ('LV',),
        'lzh': ('TW',), 'mag': ('IN',), 'mai': ('IN',), 'mg': ('MG',),
        'mhr': ('RU',), 'mi': ('NZ',), 'mk': ('MK',), 'ml': ('IN',),
        'mn': ('MN',), 'mni': ('IN',), 'mr': ('IN',), 'ms': ('MY',),
        'mt': ('MT',), 'my': ('MM',), 'nan': ('TW',), 'nb': ('NO',),
        'nds': ('DE', 'NL'), 'ne': ('NP',), 'nhn': ('MX',),
        'niu': ('NU', 'NZ'), 'nl': ('AW', 'BE', 'NL'), 'nn': ('NO',),
        'nr': ('ZA',), 'nso': ('ZA',), 'oc': ('FR',), 'om': ('ET', 'KE'),
        'or': ('IN',), 'os': ('RU',), 'pa': ('IN', 'PK'),
        'pap': ('AN', 'AW', 'CW'), 'pl': ('PL',), 'ps': ('AF',),
        'pt': ('BR', 'PT'), 'quz': ('PE',), 'raj': ('IN',), 'ro': ('RO',),
        'ru': ('RU', 'UA'), 'rw': ('RW',), 'sa': ('IN',), 'sat': ('IN',),
        'sc': ('IT',), 'sd': ('IN', 'PK'), 'se': ('NO',), 'shs': ('CA',),
        'si': ('LK',), 'sid': ('ET',), 'sk': ('SK',), 'sl': ('SI',),
        'so': ('DJ', 'ET', 'KE', 'SO'), 'sq': ('AL', 'ML'), 'sr': ('ME', 'RS'),
        'ss': ('ZA',), 'st': ('ZA',), 'sv': ('FI', 'SE'), 'sw': ('KE', 'TZ'),
        'szl': ('PL',), 'ta': ('IN', 'LK'), 'tcy': ('IN',), 'te': ('IN',),
        'tg': ('TJ',), 'th': ('TH',), 'the': ('NP',), 'ti': ('ER', 'ET'),
        'tig': ('ER',), 'tk': ('TM',), 'tl': ('PH',), 'tn': ('ZA',),
        'tr': ('CY', 'TR'), 'ts': ('ZA',), 'tt': ('RU',), 'ug': ('CN',),
        'uk': ('UA',), 'unm': ('US',), 'ur': ('IN', 'PK'), 'uz': ('UZ',),
        've': ('ZA',), 'vi': ('VN',), 'wa': ('BE',), 'wae': ('CH',),
        'wal': ('ET',), 'wo': ('SN',), 'xh': ('ZA',), 'yi': ('US',),
        'yo': ('NG',), 'yue': ('HK',), 'zh': ('CN', 'HK', 'SG', 'TW'),
        'zu': ('ZA',),
    }

    def __init__(self, generator):
        self.generator = generator

    def locale(self):
        """Generate a random underscored i18n locale code (e.g. en_US).

        :sample:
        """
        language_code = self.language_code()
        return language_code + '_' + self.random_element(
            BaseProvider.language_locale_codes[language_code],
        )

    def language_code(self):
        """Generate a random i18n language code (e.g. en).

        :sample:
        """
        return self.random_element(BaseProvider.language_locale_codes.keys())

    def random_int(self, min=0, max=9999, step=1):
        """Generate a random integer between two integers ``min`` and ``max`` inclusive
        while observing the provided ``step`` value.

        This method is functionally equivalent to randomly sampling an integer
        from the sequence ``range(min, max + 1, step)``.

        :sample:
        :sample size=10: min=0, max=15
        :sample size=10: min=0, max=15, step=3
        """
        return self.generator.random.randrange(min, max + 1, step)

    def random_digit(self):
        """Generate a random digit (0 to 9).

        :sample:
        """
        return self.generator.random.randint(0, 9)

    def random_digit_not_null(self):
        """Generate a random non-zero digit (1 to 9).

        :sample:
        """
        return self.generator.random.randint(1, 9)

    def random_digit_or_empty(self):
        """Generate a random digit (0 to 9) or an empty string.

        This method will return an empty string 50% of the time,
        and each digit has a 1/20 chance of being generated.

        :sample size=10:
        """
        if self.generator.random.randint(0, 1):
            return self.generator.random.randint(0, 9)
        else:
            return ''

    def random_digit_not_null_or_empty(self):
        """Generate a random non-zero digit (1 to 9) or an empty string.

        This method will return an empty string 50% of the time,
        and each digit has a 1/18 chance of being generated.

        :sample size=10:
        """
        if self.generator.random.randint(0, 1):
            return self.generator.random.randint(1, 9)
        else:
            return ''

    def random_number(self, digits=None, fix_len=False):
        """Generate a random integer according to the following rules:

        - If ``digits`` is ``None`` (default), its value will be set to a random
          integer from 1 to 9.
        - If ``fix_len`` is ``False`` (default), all integers that do not exceed
          the number of ``digits`` can be generated.
        - If ``fix_len`` is ``True``, only integers with the exact number of
          ``digits`` can be generated.

        :sample: fix_len=False
        :sample: fix_len=True
        :sample: digits=3
        :sample: digits=3, fix_len=False
        :sample: digits=3, fix_len=True
        """
        if digits is None:
            digits = self.random_digit_not_null()
        if digits < 0:
            raise ValueError("The digit parameter must be greater than or equal to 0.")
        if fix_len:
            if digits > 0:
                return self.generator.random.randint(
                    pow(10, digits - 1), pow(10, digits) - 1)
            else:
                raise ValueError("A number of fixed length cannot have less than 1 digit in it.")
        else:
            return self.generator.random.randint(0, pow(10, digits) - 1)

    def random_letter(self):
        """Generate a random ASCII letter (a-z and A-Z).

        :sample:
        """
        return self.generator.random.choice(
            getattr(string, 'letters', string.ascii_letters))

    def random_letters(self, length=16):
        """Generate a list of random ASCII letters (a-z and A-Z) of the specified ``length``.

        :sample:
        :sample: length=10
        """
        return self.random_choices(
            getattr(string, 'letters', string.ascii_letters),
            length=length,
        )

    def random_lowercase_letter(self):
        """Generate a random lowercase ASCII letter (a-z).

        :sample:
        """
        return self.generator.random.choice(string.ascii_lowercase)

    def random_uppercase_letter(self):
        """Generate a random uppercase ASCII letter (A-Z).

        :sample:
        """
        return self.generator.random.choice(string.ascii_uppercase)

    def random_elements(self, elements=('a', 'b', 'c'), length=None, unique=False):
        """Generate a list of randomly sampled objects from ``elements``.

        Set ``unique`` to ``False`` for random sampling with replacement, and set ``unique`` to
        ``True`` for random sampling without replacement.

        If ``length`` is set to ``None`` or is omitted, ``length`` will be set to a random
        integer from 1 to the size of ``elements``.

        The value of ``length`` cannot be greater than the number of objects
        in ``elements`` if ``unique`` is set to ``True``.

        The value of ``elements`` can be any sequence type (``list``, ``tuple``, ``set``,
        ``string``, etc) or an ``OrderedDict`` type. If it is the latter, the keys will be
        used as the objects for sampling, and the values will be used as weighted probabilities
        if ``unique`` is set to ``False``. For example:

        .. code-block:: python

            # Random sampling with replacement
            fake.random_elements(
                elements=OrderedDict([
                    ("variable_1", 0.5),        # Generates "variable_1" 50% of the time
                    ("variable_2", 0.2),        # Generates "variable_2" 20% of the time
                    ("variable_3", 0.2),        # Generates "variable_3" 20% of the time
                    ("variable_4": 0.1),        # Generates "variable_4" 10% of the time
                ]), unique=False
            )

            # Random sampling without replacement (defaults to uniform distribution)
            fake.random_elements(
                elements=OrderedDict([
                    ("variable_1", 0.5),
                    ("variable_2", 0.2),
                    ("variable_3", 0.2),
                    ("variable_4": 0.1),
                ]), unique=True
            )

        :sample: elements=('a', 'b', 'c', 'd'), unique=False
        :sample: elements=('a', 'b', 'c', 'd'), unique=True
        :sample: elements=('a', 'b', 'c', 'd'), length=10, unique=False
        :sample: elements=('a', 'b', 'c', 'd'), length=4, unique=True
        :sample: elements=OrderedDict([
                        ("a", 0.45),
                        ("b", 0.35),
                       ("c", 0.15),
                       ("d", 0.05),
                   ]), length=20, unique=False
        :sample: elements=OrderedDict([
                       ("a", 0.45),
                       ("b", 0.35),
                       ("c", 0.15),
                       ("d", 0.05),
                   ]), unique=True
        """
        if isinstance(elements, dict) and not isinstance(elements, OrderedDict):
            raise ValueError("Use OrderedDict only to avoid dependency on PYTHONHASHSEED (See #363).")

        fn = choices_distribution_unique if unique else choices_distribution

        if length is None:
            length = self.generator.random.randint(1, len(elements))

        if unique and length > len(elements):
            raise ValueError(
                "Sample length cannot be longer than the number of unique elements to pick from.")

        if isinstance(elements, dict):
            choices = elements.keys()
            probabilities = elements.values()
        else:
            if unique:
                # shortcut
                return self.generator.random.sample(elements, length)
            choices = elements
            probabilities = [1.0 for _ in range(len(choices))]

        return fn(
            list(choices),
            list(probabilities),
            self.generator.random,
            length=length,
        )

    def random_choices(self, elements=('a', 'b', 'c'), length=None):
        """Generate a list of objects randomly sampled from ``elements`` with replacement.

        For information on the ``elements`` and ``length`` arguments, please refer to
        :meth:`random_elements() <faker.providers.BaseProvider.random_elements>` which
        is used under the hood with the ``unique`` argument explicitly set to ``False``.

        :sample: elements=('a', 'b', 'c', 'd')
        :sample: elements=('a', 'b', 'c', 'd'), length=10
        :sample: elements=OrderedDict([
                     ("a", 0.45),
                     ("b", 0.35),
                     ("c", 0.15),
                     ("d", 0.05),
                 ])
        :sample: elements=OrderedDict([
                     ("a", 0.45),
                     ("b", 0.35),
                     ("c", 0.15),
                     ("d", 0.05),
                 ]), length=20
        """
        return self.random_elements(elements, length, unique=False)

    def random_element(self, elements=('a', 'b', 'c')):
        """Generate a randomly sampled object from ``elements``.

        For information on the ``elements`` argument, please refer to
        :meth:`random_elements() <faker.providers.BaseProvider.random_elements>` which
        is used under the hood with the ``unique`` argument set to ``False`` and the
        ``length`` argument set to ``1``.

        :sample: elements=('a', 'b', 'c', 'd')
        :sample size=10: elements=OrderedDict([
                     ("a", 0.45),
                     ("b", 0.35),
                     ("c", 0.15),
                     ("d", 0.05),
                 ])
        """

        return self.random_elements(elements, length=1)[0]

    def random_sample(self, elements=('a', 'b', 'c'), length=None):
        """Generate a list of objects randomly sampled from ``elements`` without replacement.

        For information on the ``elements`` and ``length`` arguments, please refer to
        :meth:`random_elements() <faker.providers.BaseProvider.random_elements>` which
        is used under the hood with the ``unique`` argument explicitly set to ``True``.

        :sample: elements=('a', 'b', 'c', 'd', 'e', 'f')
        :sample: elements=('a', 'b', 'c', 'd', 'e', 'f'), length=3
        """
        return self.random_elements(elements, length, unique=True)

    def randomize_nb_elements(
            self,
            number=10,
            le=False,
            ge=False,
            min=None,
            max=None):
        """Generate a random integer near ``number`` according to the following rules:

        - If ``le`` is ``False`` (default), allow generation up to 140% of ``number``.
          If ``True``, upper bound generation is capped at 100%.
        - If ``ge`` is ``False`` (default), allow generation down to 60% of ``number``.
          If ``True``, lower bound generation is capped at 100%.
        - If a numerical value for ``min`` is provided, generated values less than ``min``
          will be clamped at ``min``.
        - If a numerical value for ``max`` is provided, generated values greater than
          ``max`` will be clamped at ``max``.
        - If both ``le`` and ``ge`` are ``True``, the value of ``number`` will automatically
          be returned, regardless of the values supplied for ``min`` and ``max``.

        :sample: number=100
        :sample: number=100, ge=True
        :sample: number=100, ge=True, min=120
        :sample: number=100, le=True
        :sample: number=100, le=True, max=80
        :sample: number=79, le=True, ge=True, min=80
        """
        if le and ge:
            return number
        _min = 100 if ge else 60
        _max = 100 if le else 140
        nb = int(number * self.generator.random.randint(_min, _max) / 100)
        if min is not None and nb < min:
            nb = min
        if max is not None and nb > max:
            nb = max
        return nb

    def numerify(self, text='###'):
        """Generate a string with each placeholder in ``text`` replaced according
        to the following rules:

        - Number signs ('#') are replaced with a random digit (0 to 9).
        - Percent signs ('%') are replaced with a random non-zero digit (1 to 9).
        - Exclamation marks ('!') are replaced with a random digit or an empty string.
        - At symbols ('@') are replaced with a random non-zero digit or an empty string.

        Under the hood, this method uses :meth:`random_digit() <faker.providers.BaseProvider.random_digit>`,
        :meth:`random_digit_not_null() <faker.providers.BaseProvider.random_digit_not_null>`,
        :meth:`random_digit_or_empty() <faker.providers.BaseProvider.random_digit_or_empty>`,
        and :meth:`random_digit_not_null_or_empty() <faker.providers.BaseProvider.random_digit_not_null_or_empty>`
        to generate the random values.

        :sample: text='Intel Core i%-%%##K vs AMD Ryzen % %%##X'
        :sample: text='!!! !!@ !@! !@@ @!! @!@ @@! @@@'
        """
        text = _re_hash.sub(
            lambda x: str(self.random_digit()),
            text)
        text = _re_perc.sub(
            lambda x: str(self.random_digit_not_null()),
            text)
        text = _re_excl.sub(
            lambda x: str(self.random_digit_or_empty()),
            text)
        text = _re_at.sub(
            lambda x: str(self.random_digit_not_null_or_empty()),
            text)
        return text

    def lexify(self, text='????', letters=string.ascii_letters):
        """Generate a string with each question mark ('?') in ``text``
        replaced with a random character from ``letters``.

        By default, ``letters`` contains all ASCII letters, uppercase and lowercase.

        :sample: text='Random Identifier: ??????????'
        :sample: text='Random Identifier: ??????????', letters='ABCDE'
        """
        return _re_qm.sub(lambda x: self.random_element(letters), text)

    def bothify(self, text='## ??', letters=string.ascii_letters):
        """Generate a string with each placeholder in ``text`` replaced according
        to the following rules:

        - Number signs ('#') are replaced with a random digit (0 to 9).
        - Question marks ('?') are replaced with a random character from ``letters``.

        By default, ``letters`` contains all ASCII letters, uppercase and lowercase.

        Under the hood, this method uses :meth:`numerify() <faker.providers.BaseProvider.numerify>` and
        and :meth:`lexify() <faker.providers.BaseProvider.lexify>` to generate random values for number
        signs and question marks respectively.

        :sample: letters='ABCDE'
        :sample: text='Product Number: ????-########'
        :sample: text='Product Number: ????-########', letters='ABCDE'
        """
        return self.lexify(self.numerify(text), letters=letters)

    def hexify(self, text='^^^^', upper=False):
        """Generate a string with each circumflex ('^') in ``text``
        replaced with a random hexadecimal character.

        By default, ``upper`` is set to False. If set to ``True``, output
        will be formatted using uppercase hexadecimal characters.

        :sample: text='MAC Address: ^^:^^:^^:^^:^^:^^'
        :sample: text='MAC Address: ^^:^^:^^:^^:^^:^^', upper=True
        """
        letters = string.hexdigits[:-6]
        if upper:
            letters = letters.upper()
        return _re_cir.sub(lambda x: self.random_element(letters), text)



from .. import Provider as BarcodeProvider


class Provider(BarcodeProvider):
    # Source of GS1 country codes: https://gs1.org/standards/id-keys/company-prefix
    local_prefixes = (4, 5), (4, 9)

    def jan(self, length=13):
        """Generate a JAN barcode of the specified length

        Japanese local EAN barcodes are called JAN-codes https://www.dsri.jp/jan/about_jan.html

        This method is an alias and just calls
        :meth:`localized_ean()<faker.providers.barcode.ja_JP.Provider.localized_ean>` under the hood

        :sample:
        :sample: length=8
        :sample: length=13
        """
        return self.localized_ean(length)

    def jan8(self):
        """Generate a 8 digit JAN barcode

        This method is an alias and just calls
        :meth:`localized_ean8()<faker.providers.barcode.ja_JP.Provider.localized_ean8>` under the hood

        :sample:
        """
        return self.localized_ean8()

    def jan13(self):
        """Generate a 13 digit JAN barcode

        This method is an alias and just calls
        :meth:`localized_ean13()<faker.providers.barcode.ja_JP.Provider.localized_ean13>` under the hood

        :sample:
        """
        return self.localized_ean13()



"""A collection of modules for iterating through different kinds of
tree, generating tokens identical to those produced by the tokenizer
module.

To create a tree walker for a new type of tree, you need to do
implement a tree walker object (called TreeWalker by convention) that
implements a 'serialize' method taking a tree as sole argument and
returning an iterator generating tokens.
"""

from __future__ import absolute_import, division, unicode_literals

from .. import constants
from .._utils import default_etree

__all__ = ["getTreeWalker", "pprint"]

treeWalkerCache = {}


def getTreeWalker(treeType, implementation=None, **kwargs):
    """Get a TreeWalker class for various types of tree with built-in support

    :arg str treeType: the name of the tree type required (case-insensitive).
        Supported values are:

        * "dom": The xml.dom.minidom DOM implementation
        * "etree": A generic walker for tree implementations exposing an
          elementtree-like interface (known to work with ElementTree,
          cElementTree and lxml.etree).
        * "lxml": Optimized walker for lxml.etree
        * "genshi": a Genshi stream

    :arg implementation: A module implementing the tree type e.g.
        xml.etree.ElementTree or cElementTree (Currently applies to the "etree"
        tree type only).

    :arg kwargs: keyword arguments passed to the etree walker--for other
        walkers, this has no effect

    :returns: a TreeWalker class

    """

    treeType = treeType.lower()
    if treeType not in treeWalkerCache:
        if treeType == "dom":
            from . import dom
            treeWalkerCache[treeType] = dom.TreeWalker
        elif treeType == "genshi":
            from . import genshi
            treeWalkerCache[treeType] = genshi.TreeWalker
        elif treeType == "lxml":
            from . import etree_lxml
            treeWalkerCache[treeType] = etree_lxml.TreeWalker
        elif treeType == "etree":
            from . import etree
            if implementation is None:
                implementation = default_etree
            # XXX: NEVER cache here, caching is done in the etree submodule
            return etree.getETreeModule(implementation, **kwargs).TreeWalker
    return treeWalkerCache.get(treeType)


def concatenateCharacterTokens(tokens):
    pendingCharacters = []
    for token in tokens:
        type = token["type"]
        if type in ("Characters", "SpaceCharacters"):
            pendingCharacters.append(token["data"])
        else:
            if pendingCharacters:
                yield {"type": "Characters", "data": "".join(pendingCharacters)}
                pendingCharacters = []
            yield token
    if pendingCharacters:
        yield {"type": "Characters", "data": "".join(pendingCharacters)}


def pprint(walker):
    """Pretty printer for tree walkers

    Takes a TreeWalker instance and pretty prints the output of walking the tree.

    :arg walker: a TreeWalker instance

    """
    output = []
    indent = 0
    for token in concatenateCharacterTokens(walker):
        type = token["type"]
        if type in ("StartTag", "EmptyTag"):
            # tag name
            if token["namespace"] and token["namespace"] != constants.namespaces["html"]:
                if token["namespace"] in constants.prefixes:
                    ns = constants.prefixes[token["namespace"]]
                else:
                    ns = token["namespace"]
                name = "%s %s" % (ns, token["name"])
            else:
                name = token["name"]
            output.append("%s<%s>" % (" " * indent, name))
            indent += 2
            # attributes (sorted for consistent ordering)
            attrs = token["data"]
            for (namespace, localname), value in sorted(attrs.items()):
                if namespace:
                    if namespace in constants.prefixes:
                        ns = constants.prefixes[namespace]
                    else:
                        ns = namespace
                    name = "%s %s" % (ns, localname)
                else:
                    name = localname
                output.append("%s%s=\"%s\"" % (" " * indent, name, value))
            # self-closing
            if type == "EmptyTag":
                indent -= 2

        elif type == "EndTag":
            indent -= 2

        elif type == "Comment":
            output.append("%s<!-- %s -->" % (" " * indent, token["data"]))

        elif type == "Doctype":
            if token["name"]:
                if token["publicId"]:
                    output.append("""%s<!DOCTYPE %s "%s" "%s">""" %
                                  (" " * indent,
                                   token["name"],
                                   token["publicId"],
                                   token["systemId"] if token["systemId"] else ""))
                elif token["systemId"]:
                    output.append("""%s<!DOCTYPE %s "" "%s">""" %
                                  (" " * indent,
                                   token["name"],
                                   token["systemId"]))
                else:
                    output.append("%s<!DOCTYPE %s>" % (" " * indent,
                                                       token["name"]))
            else:
                output.append("%s<!DOCTYPE >" % (" " * indent,))

        elif type == "Characters":
            output.append("%s\"%s\"" % (" " * indent, token["data"]))

        elif type == "SpaceCharacters":
            assert False, "concatenateCharacterTokens should have got rid of all Space tokens"

        else:
            raise ValueError("Unknown token type, %s" % type)

    return "\n".join(output)



import gc
import os
import shutil
import time

from django.core.management.base import BaseCommand

from ... import get_asset_model


# http://djangosnippets.org/snippets/1949/ + IndexError fix
def queryset_iterator(queryset, chunksize=1000):
    """''
    Iterate over a Django Queryset ordered by the primary key

    This method loads a maximum of chunksize (default: 1000) rows in it's
    memory at the same time while django normally would load all rows in it's
    memory. Using the iterator() method only causes it to not preload all the
    classes.

    Note that the implementation of the iterator does not support ordered
    query sets.
    """
    pk = 0
    try:
        last_pk = queryset.order_by("-pk")[0].pk
    except IndexError:
        return
    queryset = queryset.order_by("pk")
    while pk < last_pk:
        for row in queryset.filter(pk__gt=pk)[:chunksize]:
            pk = row.pk
            yield row
        gc.collect()


class Command(BaseCommand):
    args = None
    help = (
        "Cleanup assets table by checking if the actual file exists in media directory."
    )
    ASSET_CHUNKSIZE = 1000

    def handle(self, *args, **options):
        t1 = time.time()

        # Process all assets
        qs = queryset_iterator(get_asset_model().objects.all(), self.ASSET_CHUNKSIZE)
        for asset in qs:
            try:
                with open(asset.file.path) as f:
                    pass
            except IOError:
                self.stdout.write("Deleting asset: {}\n".format(asset))
                asset.delete()
            else:
                filename = os.path.basename(asset.file.path)
                if not asset.rename_file() and filename != asset.user_filename:
                    new_name = os.path.join(
                        os.path.dirname(asset.file.path), asset.user_filename
                    )
                    shutil.move(asset.file.path, new_name)
                    asset.file = os.path.join(
                        os.path.dirname(asset.file.name), asset.user_filename
                    )
                    asset.save()

        t2 = time.time()
        self.stdout.write("do_work took %0.3f ms\n" % ((t2 - t1) * 1000.0))



from django import template, VERSION
from django.conf import settings
from django.utils.safestring import mark_safe

from .. import utils

register = template.Library()


@register.simple_tag
def render_bundle(bundle_name, extension=None, config='DEFAULT', attrs=''):
    tags = utils.get_as_tags(bundle_name, extension=extension, config=config, attrs=attrs)
    return mark_safe('\n'.join(tags))


@register.simple_tag
def webpack_static(asset_name, config='DEFAULT'):
    return utils.get_static(asset_name, config=config)


assignment_tag = register.simple_tag if VERSION >= (1, 9) else register.assignment_tag
@assignment_tag
def get_files(bundle_name, extension=None, config='DEFAULT'):
    """
    Returns all chunks in the given bundle.
    Example usage::

        {% get_files 'editor' 'css' as editor_css_chunks %}
        CKEDITOR.config.contentsCss = '{{ editor_css_chunks.0.publicPath }}';

    :param bundle_name: The name of the bundle
    :param extension: (optional) filter by extension
    :param config: (optional) the name of the configuration
    :return: a list of matching chunks
    """
    return utils.get_files(bundle_name, extension=extension, config=config)



from scarlet.cms import bundles, views
import pytest


class TestMainBundle(bundles.Bundle):
    navigation = bundles.PARENT
    main = views.ListView(display_fields=("user", "text"))

    class Meta:
        primary_model_bundle = True


class TestBundle1(TestMainBundle):
    navigation = bundles.PARENT

    class Meta:
        primary_model_bundle = True


class TestBundle2(TestMainBundle):
    navigation = bundles.PARENT
    dashboard = (
        ("main",),
        ("tv_main", "Landing Page", {"adm_tv_pk": "tv_main"}),
    )

    class Meta:
        primary_model_bundle = True


class TestBundle3(TestBundle2):
    navigation = bundles.PARENT


@pytest.fixture
def tbm():
    return TestMainBundle(
        name="test-main",
        title="Test main Title",
        title_plural="Test main Titles",
        parent=None,
        attr_on_parent=None,
        site=None,
    )


@pytest.fixture
def tb1():
    return TestBundle1(
        name="test1",
        title="Test1 Title",
        title_plural="Test1 Titles",
        parent=None,
        attr_on_parent=None,
        site=None,
    )


@pytest.fixture
def tb2():
    return TestBundle2(
        name="test2",
        title="Test2 Title",
        title_plural="Test2 Titles",
        parent=None,
        attr_on_parent=None,
        site=None,
    )


@pytest.fixture()
def tb3():
    return TestBundle3(
        name="test2",
        title="Test3 Title",
        title_plural="Test3 Titles",
        parent=None,
        attr_on_parent=None,
        site=None,
    )


def test_bundles(tb1, tb2, tb3, tbm):
    tb1.name = "test1 change"
    assert tb1.name == "test1 change"
    assert tb2.name == "test2"
    assert tbm.name == "test-main"

    tbm.title = "333"
    assert tb1.title == "Test1 Title"
    assert tb2.title == "Test2 Title"
    assert tbm.title == "333"

    tb2.dashboard = (("main"),)
    assert tb1.dashboard == ()
    assert tb2.dashboard == (("main"),)
    assert tbm.dashboard == ()
    assert tb3.dashboard == (
        ("main",),
        ("tv_main", "Landing Page", {"adm_tv_pk": "tv_main"}),
    )

    tb1.dashboard = (
        ("main",),
        ("tv_main", "Landing Page", {"adm_tv_pk": "tv_main"}),
    )
    assert tbm.dashboard == ()
    assert tb1.dashboard == (
        ("main",),
        ("tv_main", "Landing Page", {"adm_tv_pk": "tv_main"}),
    )

    assert tb2.dashboard == (("main"),)
    assert tb3.dashboard == (
        ("main",),
        ("tv_main", "Landing Page", {"adm_tv_pk": "tv_main"}),
    )

    tb3.name = "test3"
    assert tbm.name == "test-main"
    assert tb1.name == "test1 change"
    assert tb2.name == "test2"
    assert tb3.name == "test3"



from .. import BaseProvider

localized = True


class Provider(BaseProvider):
    # Source of GS1 country codes: https://gs1.org/standards/id-keys/company-prefix
    local_prefixes = ()

    def _ean(self, length=13, prefixes=()):
        if length not in (8, 13):
            raise AssertionError("length can only be 8 or 13")

        code = [self.random_digit() for _ in range(length - 1)]

        if prefixes:
            prefix = self.random_element(prefixes)
            code[:len(prefix)] = map(int, prefix)

        if length == 8:
            weights = [3, 1, 3, 1, 3, 1, 3]
        elif length == 13:
            weights = [1, 3, 1, 3, 1, 3, 1, 3, 1, 3, 1, 3]

        weighted_sum = sum(x * y for x, y in zip(code, weights))
        check_digit = (10 - weighted_sum % 10) % 10
        code.append(check_digit)

        return ''.join(str(x) for x in code)

    def ean(self, length=13, prefixes=()):
        """Generate an EAN barcode of the specified ``length``.

        The value of ``length`` can only be ``8`` or ``13`` (default)  which will
        create an EAN-8 or an EAN-13 barcode respectively.

        If ``prefixes`` are specified, the result will begin with one of the sequence in ``prefixes``

        :sample: length=13
        :sample: length=8
        :sample: prefixes=('00',)
        :sample: prefixes=('45', '49')
        """
        return self._ean(length, prefixes=prefixes)

    def ean8(self, prefixes=()):
        """Generate an EAN-8 barcode.

        This method uses :meth:`ean() <faker.providers.barcode.Provider.ean>` under the
        hood with the ``length`` argument explicitly set to ``8``.

        If ``prefixes`` are specified, the result will begin with one of the sequence in ``prefixes``

        :sample:
        :sample: prefixes=('00',)
        :sample: prefixes=('45', '49')
        """
        return self._ean(8, prefixes=prefixes)

    def ean13(self, prefixes=()):
        """Generate an EAN-13 barcode.

        If ``prefixes`` are specified, the result will begin with one of the sequence in ``prefixes``.

        This method uses :meth:`ean() <faker.providers.barcode.Provider.ean>` under the
        hood with the ``length`` argument explicitly set to ``13``.

        .. note::

            Codes starting with a leading zero are treated specially in some barcode readers.

            For more information about compatibility with UPC-A codes, see
            :meth:`en_US.ean13() <faker.providers.barcode.en_US.Provider.ean13>`

        :sample:
        :sample: prefixes=('00',)
        :sample: prefixes=('45', '49')
        """

        return self._ean(13, prefixes=prefixes)

    def localized_ean(self, length=13):
        """Generate a localized EAN barcode of the specified ``length``.

        The value of ``length`` can only be ``8`` or ``13`` (default)  which will
        create an EAN-8 or an EAN-13 barcode respectively.

        This method uses :meth:`ean() <faker.providers.barcode.Provider.ean>` under the
        hood with the ``prefixes`` argument explicitly set to ``self.local_prefixes``.

        :sample:
        :sample: length=13
        :sample: length=8
        """
        return self._ean(length, prefixes=self.local_prefixes)

    def localized_ean8(self):
        """Generate a localized EAN-8 barcode.

        This method uses :meth:`localized_ean() <faker.providers.barcode.Provider.ean>` under the
        hood with the ``length`` argument explicitly set to ``8``.

        :sample:
        """
        return self.localized_ean(8)

    def localized_ean13(self):
        """Generate a localized EAN-13 barcode.

        This method uses :meth:`localized_ean() <faker.providers.barcode.Provider.ean>` under the
        hood with the ``length`` argument explicitly set to ``13``.

        :sample:
        """
        return self.localized_ean(13)



import logging

from faker.providers.bank import Provider as BankProvider

logger = logging.getLogger(__name__)


class Provider(BankProvider):
    country_code = 'PH'
    bban_format = '################'
    swift_bank_codes = (
        'ANZB', 'AUBK', 'BKCH', 'BKKB', 'BNOR', 'BNPA', 'BOFA', 'BOPI',
        'BOTK', 'BPDI', 'BPFS', 'BPGO', 'CHAS', 'CHBK', 'CHSV', 'CITI',
        'CPHI', 'CTCB', 'DBPH', 'DEUT', 'EQSN', 'EWBC', 'FCBK', 'HBPH',
        'HNBK', 'HSBC', 'IBKO', 'ICBC', 'INGB', 'KOEX', 'MBBE', 'MBTC',
        'MHCB', 'PABI', 'PHSB', 'PHTB', 'PHVB', 'PNBM', 'PPBU', 'RCBC',
        'ROBP', 'SCBL', 'SETC', 'SHBK', 'SMBC', 'STLA', 'TACB', 'TLBP',
        'TYBK', 'UBPH', 'UCPB', 'UOVB', 'UWCB',
    )
    swift_location_codes = (
        '22', '2X', 'M1',
        'MM', 'MQ', 'MX',
    )
    swift_branch_codes = (
        'CBU', 'EQI', 'TSU', 'XXX',
    )

    def bban(self):
        """Generate a BBAN

        .. warning::

           Philippine bank accounts do not have BBANs or IBANs, so any number generated by this
           method is a purely hypothetical number. Local bank account numbers are typically 10
           or 12 digits long, so the BBAN format used in this implementation has been arbitrarily
           set to 16 digits to simulate a hypothetical standardization of account numbers. Using
           this method will also log a warning regarding the hypotheticality of the result.

        :sample:
        """
        logger.warning("Numbers generated by this method are purely hypothetical.")
        return super().bban()

    def iban(self):
        """Generate a BBAN

        .. warning::

           Philippine bank accounts do not have BBANs or IBANs, so any number generated by this
           method is a purely hypothetical number. This method uses hypothetical PH BBANs and the
           PH country code as inputs to the IBAN generation algorithm. Using this method will also
           log a warning regarding the hypotheticality of the result.

        :sample:
        """
        logger.warning("Numbers generated by this method are purely hypothetical.")
        return super().iban()



from django.db.migrations.serializer import BaseSerializer


class RangeSerializer(BaseSerializer):
    def serialize(self):
        module = self.value.__class__.__module__
        # Ranges are implemented in psycopg2._range but the public import
        # location is psycopg2.extras.
        module = 'psycopg2.extras' if module == 'psycopg2._range' else module
        return '%s.%r' % (module, self.value), {'import %s' % module}



from collections import defaultdict, namedtuple

from django.contrib.gis import forms, gdal
from django.contrib.gis.db.models.proxy import SpatialProxy
from django.contrib.gis.gdal.error import GDALException
from django.contrib.gis.geos import (
    GeometryCollection, GEOSException, GEOSGeometry, LineString,
    MultiLineString, MultiPoint, MultiPolygon, Point, Polygon,
)
from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields import Field
from django.utils.translation import gettext_lazy as _

# Local cache of the spatial_ref_sys table, which holds SRID data for each
# spatial database alias. This cache exists so that the database isn't queried
# for SRID info each time a distance query is constructed.
_srid_cache = defaultdict(dict)


SRIDCacheEntry = namedtuple('SRIDCacheEntry', ['units', 'units_name', 'spheroid', 'geodetic'])


def get_srid_info(srid, connection):
    """
    Return the units, unit name, and spheroid WKT associated with the
    given SRID from the `spatial_ref_sys` (or equivalent) spatial database
    table for the given database connection.  These results are cached.
    """
    from django.contrib.gis.gdal import SpatialReference
    global _srid_cache

    try:
        # The SpatialRefSys model for the spatial backend.
        SpatialRefSys = connection.ops.spatial_ref_sys()
    except NotImplementedError:
        SpatialRefSys = None

    alias, get_srs = (
        (connection.alias, lambda srid: SpatialRefSys.objects.using(connection.alias).get(srid=srid).srs)
        if SpatialRefSys else
        (None, SpatialReference)
    )
    if srid not in _srid_cache[alias]:
        srs = get_srs(srid)
        units, units_name = srs.units
        _srid_cache[alias][srid] = SRIDCacheEntry(
            units=units,
            units_name=units_name,
            spheroid='SPHEROID["%s",%s,%s]' % (srs['spheroid'], srs.semi_major, srs.inverse_flattening),
            geodetic=srs.geographic,
        )

    return _srid_cache[alias][srid]


class BaseSpatialField(Field):
    """
    The Base GIS Field.

    It's used as a base class for GeometryField and RasterField. Defines
    properties that are common to all GIS fields such as the characteristics
    of the spatial reference system of the field.
    """
    description = _("The base GIS field.")
    empty_strings_allowed = False

    def __init__(self, verbose_name=None, srid=4326, spatial_index=True, **kwargs):
        """
        The initialization function for base spatial fields. Takes the following
        as keyword arguments:

        srid:
         The spatial reference system identifier, an OGC standard.
         Defaults to 4326 (WGS84).

        spatial_index:
         Indicates whether to create a spatial index.  Defaults to True.
         Set this instead of 'db_index' for geographic fields since index
         creation is different for geometry columns.
        """

        # Setting the index flag with the value of the `spatial_index` keyword.
        self.spatial_index = spatial_index

        # Setting the SRID and getting the units.  Unit information must be
        # easily available in the field instance for distance queries.
        self.srid = srid

        # Setting the verbose_name keyword argument with the positional
        # first parameter, so this works like normal fields.
        kwargs['verbose_name'] = verbose_name

        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # Always include SRID for less fragility; include spatial index if it's
        # not the default value.
        kwargs['srid'] = self.srid
        if self.spatial_index is not True:
            kwargs['spatial_index'] = self.spatial_index
        return name, path, args, kwargs

    def db_type(self, connection):
        return connection.ops.geo_db_type(self)

    def spheroid(self, connection):
        return get_srid_info(self.srid, connection).spheroid

    def units(self, connection):
        return get_srid_info(self.srid, connection).units

    def units_name(self, connection):
        return get_srid_info(self.srid, connection).units_name

    def geodetic(self, connection):
        """
        Return true if this field's SRID corresponds with a coordinate
        system that uses non-projected units (e.g., latitude/longitude).
        """
        return get_srid_info(self.srid, connection).geodetic

    def get_placeholder(self, value, compiler, connection):
        """
        Return the placeholder for the spatial column for the
        given value.
        """
        return connection.ops.get_geom_placeholder(self, value, compiler)

    def get_srid(self, obj):
        """
        Return the default SRID for the given geometry or raster, taking into
        account the SRID set for the field. For example, if the input geometry
        or raster doesn't have an SRID, then the SRID of the field will be
        returned.
        """
        srid = obj.srid  # SRID of given geometry.
        if srid is None or self.srid == -1 or (srid == -1 and self.srid != -1):
            return self.srid
        else:
            return srid

    def get_db_prep_value(self, value, connection, *args, **kwargs):
        if value is None:
            return None
        return connection.ops.Adapter(
            super().get_db_prep_value(value, connection, *args, **kwargs),
            **({'geography': True} if self.geography and connection.ops.geography else {})
        )

    def get_raster_prep_value(self, value, is_candidate):
        """
        Return a GDALRaster if conversion is successful, otherwise return None.
        """
        if isinstance(value, gdal.GDALRaster):
            return value
        elif is_candidate:
            try:
                return gdal.GDALRaster(value)
            except GDALException:
                pass
        elif isinstance(value, dict):
            try:
                return gdal.GDALRaster(value)
            except GDALException:
                raise ValueError("Couldn't create spatial object from lookup value '%s'." % value)

    def get_prep_value(self, value):
        obj = super().get_prep_value(value)
        if obj is None:
            return None
        # When the input is not a geometry or raster, attempt to construct one
        # from the given string input.
        if isinstance(obj, GEOSGeometry):
            pass
        else:
            # Check if input is a candidate for conversion to raster or geometry.
            is_candidate = isinstance(obj, (bytes, str)) or hasattr(obj, '__geo_interface__')
            # Try to convert the input to raster.
            raster = self.get_raster_prep_value(obj, is_candidate)

            if raster:
                obj = raster
            elif is_candidate:
                try:
                    obj = GEOSGeometry(obj)
                except (GEOSException, GDALException):
                    raise ValueError("Couldn't create spatial object from lookup value '%s'." % obj)
            else:
                raise ValueError('Cannot use object with type %s for a spatial lookup parameter.' % type(obj).__name__)

        # Assigning the SRID value.
        obj.srid = self.get_srid(obj)
        return obj


class GeometryField(BaseSpatialField):
    """
    The base Geometry field -- maps to the OpenGIS Specification Geometry type.
    """
    description = _("The base Geometry field -- maps to the OpenGIS Specification Geometry type.")
    form_class = forms.GeometryField
    # The OpenGIS Geometry name.
    geom_type = 'GEOMETRY'
    geom_class = None

    def __init__(self, verbose_name=None, dim=2, geography=False, *, extent=(-180.0, -90.0, 180.0, 90.0),
                 tolerance=0.05, **kwargs):
        """
        The initialization function for geometry fields. In addition to the
        parameters from BaseSpatialField, it takes the following as keyword
        arguments:

        dim:
         The number of dimensions for this geometry.  Defaults to 2.

        extent:
         Customize the extent, in a 4-tuple of WGS 84 coordinates, for the
         geometry field entry in the `USER_SDO_GEOM_METADATA` table.  Defaults
         to (-180.0, -90.0, 180.0, 90.0).

        tolerance:
         Define the tolerance, in meters, to use for the geometry field
         entry in the `USER_SDO_GEOM_METADATA` table.  Defaults to 0.05.
        """
        # Setting the dimension of the geometry field.
        self.dim = dim

        # Is this a geography rather than a geometry column?
        self.geography = geography

        # Oracle-specific private attributes for creating the entry in
        # `USER_SDO_GEOM_METADATA`
        self._extent = extent
        self._tolerance = tolerance

        super().__init__(verbose_name=verbose_name, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # Include kwargs if they're not the default values.
        if self.dim != 2:
            kwargs['dim'] = self.dim
        if self.geography is not False:
            kwargs['geography'] = self.geography
        if self._extent != (-180.0, -90.0, 180.0, 90.0):
            kwargs['extent'] = self._extent
        if self._tolerance != 0.05:
            kwargs['tolerance'] = self._tolerance
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)

        # Setup for lazy-instantiated Geometry object.
        setattr(cls, self.attname, SpatialProxy(self.geom_class or GEOSGeometry, self, load_func=GEOSGeometry))

    def formfield(self, **kwargs):
        defaults = {
            'form_class': self.form_class,
            'geom_type': self.geom_type,
            'srid': self.srid,
            **kwargs,
        }
        if self.dim > 2 and not getattr(defaults['form_class'].widget, 'supports_3d', False):
            defaults.setdefault('widget', forms.Textarea)
        return super().formfield(**defaults)

    def select_format(self, compiler, sql, params):
        """
        Return the selection format string, depending on the requirements
        of the spatial backend. For example, Oracle and MySQL require custom
        selection formats in order to retrieve geometries in OGC WKB.
        """
        return compiler.connection.ops.select % sql, params


# The OpenGIS Geometry Type Fields
class PointField(GeometryField):
    geom_type = 'POINT'
    geom_class = Point
    form_class = forms.PointField
    description = _("Point")


class LineStringField(GeometryField):
    geom_type = 'LINESTRING'
    geom_class = LineString
    form_class = forms.LineStringField
    description = _("Line string")


class PolygonField(GeometryField):
    geom_type = 'POLYGON'
    geom_class = Polygon
    form_class = forms.PolygonField
    description = _("Polygon")


class MultiPointField(GeometryField):
    geom_type = 'MULTIPOINT'
    geom_class = MultiPoint
    form_class = forms.MultiPointField
    description = _("Multi-point")


class MultiLineStringField(GeometryField):
    geom_type = 'MULTILINESTRING'
    geom_class = MultiLineString
    form_class = forms.MultiLineStringField
    description = _("Multi-line string")


class MultiPolygonField(GeometryField):
    geom_type = 'MULTIPOLYGON'
    geom_class = MultiPolygon
    form_class = forms.MultiPolygonField
    description = _("Multi polygon")


class GeometryCollectionField(GeometryField):
    geom_type = 'GEOMETRYCOLLECTION'
    geom_class = GeometryCollection
    form_class = forms.GeometryCollectionField
    description = _("Geometry collection")


class ExtentField(Field):
    "Used as a return value from an extent aggregate"

    description = _("Extent Aggregate Field")

    def get_internal_type(self):
        return "ExtentField"

    def select_format(self, compiler, sql, params):
        select = compiler.connection.ops.select_extent
        return select % sql if select else sql, params


class RasterField(BaseSpatialField):
    """
    Raster field for GeoDjango -- evaluates into GDALRaster objects.
    """

    description = _("Raster Field")
    geom_type = 'RASTER'
    geography = False

    def _check_connection(self, connection):
        # Make sure raster fields are used only on backends with raster support.
        if not connection.features.gis_enabled or not connection.features.supports_raster:
            raise ImproperlyConfigured('Raster fields require backends with raster support.')

    def db_type(self, connection):
        self._check_connection(connection)
        return super().db_type(connection)

    def from_db_value(self, value, expression, connection):
        return connection.ops.parse_raster(value)

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        # Setup for lazy-instantiated Raster object. For large querysets, the
        # instantiation of all GDALRasters can potentially be expensive. This
        # delays the instantiation of the objects to the moment of evaluation
        # of the raster attribute.
        setattr(cls, self.attname, SpatialProxy(gdal.GDALRaster, self))

    def get_transform(self, name):
        from django.contrib.gis.db.models.lookups import RasterBandTransform
        try:
            band_index = int(name)
            return type(
                'SpecificRasterBandTransform',
                (RasterBandTransform,),
                {'band_index': band_index}
            )
        except ValueError:
            pass
        return super().get_transform(name)



from inspect import getfullargspec

from django.template.library import InclusionNode, parse_bits


class InclusionAdminNode(InclusionNode):
    """
    Template tag that allows its template to be overridden per model, per app,
    or globally.
    """

    def __init__(self, parser, token, func, template_name, takes_context=True):
        self.template_name = template_name
        params, varargs, varkw, defaults, kwonly, kwonly_defaults, _ = getfullargspec(func)
        bits = token.split_contents()
        args, kwargs = parse_bits(
            parser, bits[1:], params, varargs, varkw, defaults, kwonly,
            kwonly_defaults, takes_context, bits[0],
        )
        super().__init__(func, takes_context, args, kwargs, filename=None)

    def render(self, context):
        opts = context['opts']
        app_label = opts.app_label.lower()
        object_name = opts.object_name.lower()
        # Load template for this render call. (Setting self.filename isn't
        # thread-safe.)
        context.render_context[self] = context.template.engine.select_template([
            'admin/%s/%s/%s' % (app_label, object_name, self.template_name),
            'admin/%s/%s' % (app_label, self.template_name),
            'admin/%s' % (self.template_name,),
        ])
        return super().render(context)



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

import re

from sqlparse import sql, tokens as T


# FIXME: Doesn't work
class RightMarginFilter:
    keep_together = (
        # sql.TypeCast, sql.Identifier, sql.Alias,
    )

    def __init__(self, width=79):
        self.width = width
        self.line = ''

    def _process(self, group, stream):
        for token in stream:
            if token.is_whitespace and '\n' in token.value:
                if token.value.endswith('\n'):
                    self.line = ''
                else:
                    self.line = token.value.splitlines()[-1]
            elif token.is_group and type(token) not in self.keep_together:
                token.tokens = self._process(token, token.tokens)
            else:
                val = str(token)
                if len(self.line) + len(val) > self.width:
                    match = re.search(r'^ +', self.line)
                    if match is not None:
                        indent = match.group()
                    else:
                        indent = ''
                    yield sql.Token(T.Whitespace, '\n{}'.format(indent))
                    self.line = indent
                self.line += val
            yield token

    def process(self, group):
        # return
        # group.tokens = self._process(group, group.tokens)
        raise NotImplementedError



from .. import Provider as AutomotiveProvider


class Provider(AutomotiveProvider):
    # Source (english): https://en.wikipedia.org/wiki/Vehicle_registration_plates_of_France
    license_formats = (
        # New format
        '??-###-??',
        # Old format for plates < 2009
        '###-???-##',
    )



from django.db.models import CharField, Field, FloatField, TextField
from django.db.models.expressions import CombinedExpression, Func, Value
from django.db.models.functions import Cast, Coalesce
from django.db.models.lookups import Lookup


class SearchVectorExact(Lookup):
    lookup_name = 'exact'

    def process_rhs(self, qn, connection):
        if not hasattr(self.rhs, 'resolve_expression'):
            config = getattr(self.lhs, 'config', None)
            self.rhs = SearchQuery(self.rhs, config=config)
        rhs, rhs_params = super().process_rhs(qn, connection)
        return rhs, rhs_params

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s @@ %s = true' % (lhs, rhs), params


class SearchVectorField(Field):

    def db_type(self, connection):
        return 'tsvector'


class SearchQueryField(Field):

    def db_type(self, connection):
        return 'tsquery'


class SearchVectorCombinable:
    ADD = '||'

    def _combine(self, other, connector, reversed):
        if not isinstance(other, SearchVectorCombinable) or not self.config == other.config:
            raise TypeError('SearchVector can only be combined with other SearchVectors')
        if reversed:
            return CombinedSearchVector(other, connector, self, self.config)
        return CombinedSearchVector(self, connector, other, self.config)


class SearchVector(SearchVectorCombinable, Func):
    function = 'to_tsvector'
    arg_joiner = " || ' ' || "
    output_field = SearchVectorField()
    config = None

    def __init__(self, *expressions, **extra):
        super().__init__(*expressions, **extra)
        self.config = self.extra.get('config', self.config)
        weight = self.extra.get('weight')
        if weight is not None and not hasattr(weight, 'resolve_expression'):
            weight = Value(weight)
        self.weight = weight

    def resolve_expression(self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False):
        resolved = super().resolve_expression(query, allow_joins, reuse, summarize, for_save)
        if self.config:
            if not hasattr(self.config, 'resolve_expression'):
                resolved.config = Value(self.config).resolve_expression(query, allow_joins, reuse, summarize, for_save)
            else:
                resolved.config = self.config.resolve_expression(query, allow_joins, reuse, summarize, for_save)
        return resolved

    def as_sql(self, compiler, connection, function=None, template=None):
        clone = self.copy()
        clone.set_source_expressions([
            Coalesce(
                expression
                if isinstance(expression.output_field, (CharField, TextField))
                else Cast(expression, TextField()),
                Value('')
            ) for expression in clone.get_source_expressions()
        ])
        config_params = []
        if template is None:
            if clone.config:
                config_sql, config_params = compiler.compile(clone.config)
                template = '%(function)s({}::regconfig, %(expressions)s)'.format(config_sql.replace('%', '%%'))
            else:
                template = clone.template
        sql, params = super(SearchVector, clone).as_sql(compiler, connection, function=function, template=template)
        extra_params = []
        if clone.weight:
            weight_sql, extra_params = compiler.compile(clone.weight)
            sql = 'setweight({}, {})'.format(sql, weight_sql)
        return sql, config_params + params + extra_params


class CombinedSearchVector(SearchVectorCombinable, CombinedExpression):
    def __init__(self, lhs, connector, rhs, config, output_field=None):
        self.config = config
        super().__init__(lhs, connector, rhs, output_field)


class SearchQueryCombinable:
    BITAND = '&&'
    BITOR = '||'

    def _combine(self, other, connector, reversed):
        if not isinstance(other, SearchQueryCombinable):
            raise TypeError(
                'SearchQuery can only be combined with other SearchQuerys, '
                'got {}.'.format(type(other))
            )
        if reversed:
            return CombinedSearchQuery(other, connector, self, self.config)
        return CombinedSearchQuery(self, connector, other, self.config)

    # On Combinable, these are not implemented to reduce confusion with Q. In
    # this case we are actually (ab)using them to do logical combination so
    # it's consistent with other usage in Django.
    def __or__(self, other):
        return self._combine(other, self.BITOR, False)

    def __ror__(self, other):
        return self._combine(other, self.BITOR, True)

    def __and__(self, other):
        return self._combine(other, self.BITAND, False)

    def __rand__(self, other):
        return self._combine(other, self.BITAND, True)


class SearchQuery(SearchQueryCombinable, Value):
    output_field = SearchQueryField()
    SEARCH_TYPES = {
        'plain': 'plainto_tsquery',
        'phrase': 'phraseto_tsquery',
        'raw': 'to_tsquery',
    }

    def __init__(self, value, output_field=None, *, config=None, invert=False, search_type='plain'):
        self.config = config
        self.invert = invert
        if search_type not in self.SEARCH_TYPES:
            raise ValueError("Unknown search_type argument '%s'." % search_type)
        self.search_type = search_type
        super().__init__(value, output_field=output_field)

    def resolve_expression(self, query=None, allow_joins=True, reuse=None, summarize=False, for_save=False):
        resolved = super().resolve_expression(query, allow_joins, reuse, summarize, for_save)
        if self.config:
            if not hasattr(self.config, 'resolve_expression'):
                resolved.config = Value(self.config).resolve_expression(query, allow_joins, reuse, summarize, for_save)
            else:
                resolved.config = self.config.resolve_expression(query, allow_joins, reuse, summarize, for_save)
        return resolved

    def as_sql(self, compiler, connection):
        params = [self.value]
        function = self.SEARCH_TYPES[self.search_type]
        if self.config:
            config_sql, config_params = compiler.compile(self.config)
            template = '{}({}::regconfig, %s)'.format(function, config_sql)
            params = config_params + [self.value]
        else:
            template = '{}(%s)'.format(function)
        if self.invert:
            template = '!!({})'.format(template)
        return template, params

    def _combine(self, other, connector, reversed):
        combined = super()._combine(other, connector, reversed)
        combined.output_field = SearchQueryField()
        return combined

    def __invert__(self):
        return type(self)(self.value, config=self.config, invert=not self.invert)

    def __str__(self):
        result = super().__str__()
        return ('~%s' % result) if self.invert else result


class CombinedSearchQuery(SearchQueryCombinable, CombinedExpression):
    def __init__(self, lhs, connector, rhs, config, output_field=None):
        self.config = config
        super().__init__(lhs, connector, rhs, output_field)

    def __str__(self):
        return '(%s)' % super().__str__()


class SearchRank(Func):
    function = 'ts_rank'
    output_field = FloatField()

    def __init__(self, vector, query, **extra):
        if not hasattr(vector, 'resolve_expression'):
            vector = SearchVector(vector)
        if not hasattr(query, 'resolve_expression'):
            query = SearchQuery(query)
        weights = extra.get('weights')
        if weights is not None and not hasattr(weights, 'resolve_expression'):
            weights = Value(weights)
        self.weights = weights
        super().__init__(vector, query, **extra)

    def as_sql(self, compiler, connection, function=None, template=None):
        extra_params = []
        extra_context = {}
        if template is None and self.extra.get('weights'):
            if self.weights:
                template = '%(function)s(%(weights)s, %(expressions)s)'
                weight_sql, extra_params = compiler.compile(self.weights)
                extra_context['weights'] = weight_sql
        sql, params = super().as_sql(
            compiler, connection,
            function=function, template=template, **extra_context
        )
        return sql, extra_params + params


SearchVectorField.register_lookup(SearchVectorExact)


class TrigramBase(Func):
    output_field = FloatField()

    def __init__(self, expression, string, **extra):
        if not hasattr(string, 'resolve_expression'):
            string = Value(string)
        super().__init__(expression, string, **extra)


class TrigramSimilarity(TrigramBase):
    function = 'SIMILARITY'


class TrigramDistance(TrigramBase):
    function = ''
    arg_joiner = ' <-> '



from unittest import mock

from faker import Faker
import pytest
from scarlet.cache.views import CacheMixin
from scarlet.cache.router import router

CACHE = router.get_cache()
fake = Faker()


@pytest.fixture
def cache_mixin(rf):
    cache_mixin = CacheMixin()
    cache_mixin.request = rf.get("/")
    cache_mixin.request.user = mock.Mock()
    return cache_mixin


@pytest.mark.parametrize("is_staff,should_cache", [(True, False), (False, True)])
def test_should_cache_staff(is_staff, should_cache, cache_mixin):
    cache_mixin.request.user.is_staff = is_staff
    assert cache_mixin.should_cache() == should_cache


@pytest.mark.parametrize(
    "is_staff,allow_cache,should_cache",
    [
        (True, True, False),
        (True, False, False),
        (False, True, True),
        (False, False, False),
    ],
)
def test_should_cache_with_allow_cache_var(
    is_staff, allow_cache, should_cache, cache_mixin
):
    cache_mixin.request.user.is_staff = is_staff
    cache_mixin.allow_cache = allow_cache
    assert cache_mixin.should_cache() == should_cache


def test_set_do_not_cache(cache_mixin):
    cache_mixin.set_do_not_cache()
    assert not cache_mixin.allow_cache


def test_cache_version_not_implemented(cache_mixin):
    with pytest.raises(NotImplementedError):
        cache_mixin.get_cache_version()


def test_get_cache_route_group(cache_mixin):
    assert cache_mixin.get_cache_route_group() == "default"


def test_get_cache_prefix_is_ajax(cache_mixin):
    cache_mixin.request.is_ajax = lambda: True
    assert cache_mixin.get_cache_prefix() == "ajax"


def test_get_cache_prefix_not_ajax(cache_mixin):
    cache_mixin.request.is_ajax = lambda: False
    assert cache_mixin.get_cache_prefix() == ""


@mock.patch("scarlet.cache.views.settings")
def test_get_cache_prefix(mocked_settings, cache_mixin):
    test_str = "test"
    mocked_settings.CACHE_MIDDLEWARE_KEY_PREFIX = test_str
    cache_mixin.request.is_ajax = lambda: True
    assert cache_mixin.get_cache_prefix() == test_str + "ajax"


@mock.patch.object(CacheMixin, "set_cache_middleware")
def test_cache_prefix_used(mocked_set_cache_middleware, cache_mixin, rf):
    """Check that a prefix is used when caching"""
    request = rf.get("/dummy_cache_view/")
    request.user = mock.Mock()
    request.user.is_staff = False
    request.is_ajax = lambda: True

    key = fake.pystr()
    cache_mixin.get_cache_version = lambda: key

    # The cache middleware is being mocked here, so there will be a Nonetype related attribute error.
    # Just ignore it and check that the call is made
    try:
        cache_mixin.dispatch(request)
    except AttributeError:
        pass
    mocked_set_cache_middleware.assert_called_once_with(
        cache_mixin.cache_time, f"{key}:{cache_mixin.get_cache_prefix()}"
    )



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0002_logentry_remove_auto_add'),
    ]

    # No database changes; adds choices to action_flag.
    operations = [
        migrations.AlterField(
            model_name='logentry',
            name='action_flag',
            field=models.PositiveSmallIntegerField(
                choices=[(1, 'Addition'), (2, 'Change'), (3, 'Deletion')],
                verbose_name='action flag',
            ),
        ),
    ]



"""A pytest plugin which helps testing Django applications

This plugin handles creating and destroying the test environment and
test database and provides some useful text fixtures.
"""

import contextlib
import inspect
from functools import reduce
import os
import sys
import types

import pytest
from pkg_resources import parse_version

from .django_compat import is_django_unittest  # noqa
from .fixtures import django_assert_num_queries  # noqa
from .fixtures import django_assert_max_num_queries  # noqa
from .fixtures import django_db_setup  # noqa
from .fixtures import django_db_use_migrations  # noqa
from .fixtures import django_db_keepdb  # noqa
from .fixtures import django_db_createdb  # noqa
from .fixtures import django_db_modify_db_settings  # noqa
from .fixtures import django_db_modify_db_settings_parallel_suffix  # noqa
from .fixtures import django_db_modify_db_settings_tox_suffix  # noqa
from .fixtures import django_db_modify_db_settings_xdist_suffix  # noqa
from .fixtures import _live_server_helper  # noqa
from .fixtures import admin_client  # noqa
from .fixtures import admin_user  # noqa
from .fixtures import client  # noqa
from .fixtures import db  # noqa
from .fixtures import django_user_model  # noqa
from .fixtures import django_username_field  # noqa
from .fixtures import live_server  # noqa
from .fixtures import django_db_reset_sequences  # noqa
from .fixtures import rf  # noqa
from .fixtures import settings  # noqa
from .fixtures import transactional_db  # noqa

from .lazy_django import django_settings_is_configured, skip_if_no_django

try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib


SETTINGS_MODULE_ENV = "DJANGO_SETTINGS_MODULE"
CONFIGURATION_ENV = "DJANGO_CONFIGURATION"
INVALID_TEMPLATE_VARS_ENV = "FAIL_INVALID_TEMPLATE_VARS"

PY2 = sys.version_info[0] == 2

# pytest 4.2 handles unittest setup/teardown itself via wrapping fixtures.
_handle_unittest_methods = parse_version(pytest.__version__) < parse_version("4.2")

_report_header = []


# ############### pytest hooks ################


def pytest_addoption(parser):
    group = parser.getgroup("django")
    group._addoption(
        "--reuse-db",
        action="store_true",
        dest="reuse_db",
        default=False,
        help="Re-use the testing database if it already exists, "
        "and do not remove it when the test finishes.",
    )
    group._addoption(
        "--create-db",
        action="store_true",
        dest="create_db",
        default=False,
        help="Re-create the database, even if it exists. This "
        "option can be used to override --reuse-db.",
    )
    group._addoption(
        "--ds",
        action="store",
        type=str,
        dest="ds",
        default=None,
        help="Set DJANGO_SETTINGS_MODULE.",
    )
    group._addoption(
        "--dc",
        action="store",
        type=str,
        dest="dc",
        default=None,
        help="Set DJANGO_CONFIGURATION.",
    )
    group._addoption(
        "--nomigrations",
        "--no-migrations",
        action="store_true",
        dest="nomigrations",
        default=False,
        help="Disable Django migrations on test setup",
    )
    group._addoption(
        "--migrations",
        action="store_false",
        dest="nomigrations",
        default=False,
        help="Enable Django migrations on test setup",
    )
    parser.addini(
        CONFIGURATION_ENV, "django-configurations class to use by pytest-django."
    )
    group._addoption(
        "--liveserver",
        default=None,
        help="Address and port for the live_server fixture.",
    )
    parser.addini(
        SETTINGS_MODULE_ENV, "Django settings module to use by pytest-django."
    )

    parser.addini(
        "django_find_project",
        "Automatically find and add a Django project to the " "Python path.",
        type="bool",
        default=True,
    )
    group._addoption(
        "--fail-on-template-vars",
        action="store_true",
        dest="itv",
        default=False,
        help="Fail for invalid variables in templates.",
    )
    parser.addini(
        INVALID_TEMPLATE_VARS_ENV,
        "Fail for invalid variables in templates.",
        type="bool",
        default=False,
    )


PROJECT_FOUND = (
    "pytest-django found a Django project in %s "
    "(it contains manage.py) and added it to the Python path.\n"
    'If this is wrong, add "django_find_project = false" to '
    "pytest.ini and explicitly manage your Python path."
)

PROJECT_NOT_FOUND = (
    "pytest-django could not find a Django project "
    "(no manage.py file could be found). You must "
    "explicitly add your Django project to the Python path "
    "to have it picked up."
)

PROJECT_SCAN_DISABLED = (
    "pytest-django did not search for Django "
    "projects since it is disabled in the configuration "
    '("django_find_project = false")'
)


@contextlib.contextmanager
def _handle_import_error(extra_message):
    try:
        yield
    except ImportError as e:
        django_msg = (e.args[0] + "\n\n") if e.args else ""
        msg = django_msg + extra_message
        raise ImportError(msg)


def _add_django_project_to_path(args):
    def is_django_project(path):
        try:
            return path.is_dir() and (path / "manage.py").exists()
        except OSError:
            return False

    def arg_to_path(arg):
        # Test classes or functions can be appended to paths separated by ::
        arg = arg.split("::", 1)[0]
        return pathlib.Path(arg)

    def find_django_path(args):
        args = map(str, args)
        args = [arg_to_path(x) for x in args if not x.startswith("-")]

        cwd = pathlib.Path.cwd()
        if not args:
            args.append(cwd)
        elif cwd not in args:
            args.append(cwd)

        for arg in args:
            if is_django_project(arg):
                return arg
            for parent in arg.parents:
                if is_django_project(parent):
                    return parent
        return None

    project_dir = find_django_path(args)
    if project_dir:
        sys.path.insert(0, str(project_dir.absolute()))
        return PROJECT_FOUND % project_dir
    return PROJECT_NOT_FOUND


def _setup_django():
    if "django" not in sys.modules:
        return

    import django.conf

    # Avoid force-loading Django when settings are not properly configured.
    if not django.conf.settings.configured:
        return

    import django.apps

    if not django.apps.apps.ready:
        django.setup()

    _blocking_manager.block()


def _get_boolean_value(x, name, default=None):
    if x is None:
        return default
    if x in (True, False):
        return x
    possible_values = {"true": True, "false": False, "1": True, "0": False}
    try:
        return possible_values[x.lower()]
    except KeyError:
        raise ValueError(
            "{} is not a valid value for {}. "
            "It must be one of {}." % (x, name, ", ".join(possible_values.keys()))
        )


def pytest_load_initial_conftests(early_config, parser, args):
    # Register the marks
    early_config.addinivalue_line(
        "markers",
        "django_db(transaction=False): Mark the test as using "
        "the Django test database.  The *transaction* argument marks will "
        "allow you to use real transactions in the test like Django's "
        "TransactionTestCase.",
    )
    early_config.addinivalue_line(
        "markers",
        "urls(modstr): Use a different URLconf for this test, similar to "
        "the `urls` attribute of Django's `TestCase` objects.  *modstr* is "
        "a string specifying the module of a URL config, e.g. "
        '"my_app.test_urls".',
    )
    early_config.addinivalue_line(
        "markers",
        "ignore_template_errors(): ignore errors from invalid template "
        "variables (if --fail-on-template-vars is used).",
    )

    options = parser.parse_known_args(args)

    if options.version or options.help:
        return

    django_find_project = _get_boolean_value(
        early_config.getini("django_find_project"), "django_find_project"
    )

    if django_find_project:
        _django_project_scan_outcome = _add_django_project_to_path(args)
    else:
        _django_project_scan_outcome = PROJECT_SCAN_DISABLED

    if (
        options.itv
        or _get_boolean_value(
            os.environ.get(INVALID_TEMPLATE_VARS_ENV), INVALID_TEMPLATE_VARS_ENV
        )
        or early_config.getini(INVALID_TEMPLATE_VARS_ENV)
    ):
        os.environ[INVALID_TEMPLATE_VARS_ENV] = "true"

    def _get_option_with_source(option, envname):
        if option:
            return option, "option"
        if envname in os.environ:
            return os.environ[envname], "env"
        cfgval = early_config.getini(envname)
        if cfgval:
            return cfgval, "ini"
        return None, None

    ds, ds_source = _get_option_with_source(options.ds, SETTINGS_MODULE_ENV)
    dc, dc_source = _get_option_with_source(options.dc, CONFIGURATION_ENV)

    if ds:
        _report_header.append("settings: %s (from %s)" % (ds, ds_source))
        os.environ[SETTINGS_MODULE_ENV] = ds

        if dc:
            _report_header.append("configuration: %s (from %s)" % (dc, dc_source))
            os.environ[CONFIGURATION_ENV] = dc

            # Install the django-configurations importer
            import configurations.importer

            configurations.importer.install()

        # Forcefully load Django settings, throws ImportError or
        # ImproperlyConfigured if settings cannot be loaded.
        from django.conf import settings as dj_settings

        with _handle_import_error(_django_project_scan_outcome):
            dj_settings.DATABASES

    _setup_django()


def pytest_report_header():
    if _report_header:
        return ["django: " + ", ".join(_report_header)]


@pytest.mark.trylast
def pytest_configure():
    # Allow Django settings to be configured in a user pytest_configure call,
    # but make sure we call django.setup()
    _setup_django()


def _classmethod_is_defined_at_leaf(cls, method_name):
    super_method = None

    for base_cls in cls.__mro__[1:]:  # pragma: no branch
        super_method = base_cls.__dict__.get(method_name)
        if super_method is not None:
            break

    assert super_method is not None, (
        "%s could not be found in base classes" % method_name
    )

    method = getattr(cls, method_name)

    try:
        f = method.__func__
    except AttributeError:
        pytest.fail("%s.%s should be a classmethod" % (cls, method_name))
    if PY2 and not (
        inspect.ismethod(method)
        and inspect.isclass(method.__self__)
        and issubclass(cls, method.__self__)
    ):
        pytest.fail("%s.%s should be a classmethod" % (cls, method_name))
    return f is not super_method.__func__


_disabled_classmethods = {}


def _disable_class_methods(cls):
    if cls in _disabled_classmethods:
        return

    _disabled_classmethods[cls] = (
        # Get the classmethod object (not the resulting bound method),
        # otherwise inheritance will be broken when restoring.
        cls.__dict__.get("setUpClass"),
        _classmethod_is_defined_at_leaf(cls, "setUpClass"),
        cls.__dict__.get("tearDownClass"),
        _classmethod_is_defined_at_leaf(cls, "tearDownClass"),
    )

    cls.setUpClass = types.MethodType(lambda cls: None, cls)
    cls.tearDownClass = types.MethodType(lambda cls: None, cls)


def _restore_class_methods(cls):
    (
        setUpClass,
        restore_setUpClass,
        tearDownClass,
        restore_tearDownClass,
    ) = _disabled_classmethods.pop(cls)

    try:
        del cls.setUpClass
    except AttributeError:
        raise

    try:
        del cls.tearDownClass
    except AttributeError:
        pass

    if restore_setUpClass:
        cls.setUpClass = setUpClass

    if restore_tearDownClass:
        cls.tearDownClass = tearDownClass


def pytest_runtest_setup(item):
    if _handle_unittest_methods:
        if django_settings_is_configured() and is_django_unittest(item):
            _disable_class_methods(item.cls)


def pytest_collection_modifyitems(items):
    def get_order_number(test):
        marker_db = test.get_closest_marker('django_db')
        if marker_db:
            transaction = validate_django_db(marker_db)[0]
            if transaction is True:
                return 1
        else:
            transaction = None

        fixtures = getattr(test, 'fixturenames', [])
        if "transactional_db" in fixtures:
            return 1

        if transaction is False:
            return 0
        if "db" in fixtures:
            return 0

        return 2

    items[:] = sorted(items, key=get_order_number)


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(request):
    """
    Ensure that Django is loaded and has its testing environment setup.

    XXX It is a little dodgy that this is an autouse fixture.  Perhaps
        an email fixture should be requested in order to be able to
        use the Django email machinery just like you need to request a
        db fixture for access to the Django database, etc.  But
        without duplicating a lot more of Django's test support code
        we need to follow this model.
    """
    if django_settings_is_configured():
        _setup_django()
        from django.conf import settings as dj_settings
        from django.test.utils import setup_test_environment, teardown_test_environment

        dj_settings.DEBUG = False
        setup_test_environment()
        request.addfinalizer(teardown_test_environment)


@pytest.fixture(scope="session")
def django_db_blocker():
    """Wrapper around Django's database access.

    This object can be used to re-enable database access.  This fixture is used
    internally in pytest-django to build the other fixtures and can be used for
    special database handling.

    The object is a context manager and provides the methods
    .unblock()/.block() and .restore() to temporarily enable database access.

    This is an advanced feature that is meant to be used to implement database
    fixtures.
    """
    if not django_settings_is_configured():
        return None

    return _blocking_manager


@pytest.fixture(autouse=True)
def _django_db_marker(request):
    """Implement the django_db marker, internal to pytest-django.

    This will dynamically request the ``db``, ``transactional_db`` or
    ``django_db_reset_sequences`` fixtures as required by the django_db marker.
    """
    marker = request.node.get_closest_marker("django_db")
    if marker:
        transaction, reset_sequences = validate_django_db(marker)
        if reset_sequences:
            request.getfixturevalue("django_db_reset_sequences")
        elif transaction:
            request.getfixturevalue("transactional_db")
        else:
            request.getfixturevalue("db")


@pytest.fixture(autouse=True, scope="class")
def _django_setup_unittest(request, django_db_blocker):
    """Setup a django unittest, internal to pytest-django."""
    if not django_settings_is_configured() or not is_django_unittest(request):
        yield
        return

    from _pytest.unittest import TestCaseFunction

    if "debug" in TestCaseFunction.runtest.__code__.co_names:
        # Fix pytest (https://github.com/pytest-dev/pytest/issues/5991), only
        # if "self._testcase.debug()" is being used (forward compatible).
        from _pytest.monkeypatch import MonkeyPatch

        def non_debugging_runtest(self):
            self._testcase(result=self)

        mp_debug = MonkeyPatch()
        mp_debug.setattr("_pytest.unittest.TestCaseFunction.runtest", non_debugging_runtest)
    else:
        mp_debug = None

    request.getfixturevalue("django_db_setup")

    cls = request.node.cls

    with django_db_blocker.unblock():
        if _handle_unittest_methods:
            _restore_class_methods(cls)
            cls.setUpClass()
            _disable_class_methods(cls)

            yield

            _restore_class_methods(cls)
            cls.tearDownClass()
        else:
            yield

    if mp_debug:
        mp_debug.undo()


@pytest.fixture(scope="function", autouse=True)
def _dj_autoclear_mailbox():
    if not django_settings_is_configured():
        return

    from django.core import mail

    del mail.outbox[:]


@pytest.fixture(scope="function")
def mailoutbox(django_mail_patch_dns, _dj_autoclear_mailbox):
    if not django_settings_is_configured():
        return

    from django.core import mail

    return mail.outbox


@pytest.fixture(scope="function")
def django_mail_patch_dns(monkeypatch, django_mail_dnsname):
    from django.core import mail

    monkeypatch.setattr(mail.message, "DNS_NAME", django_mail_dnsname)


@pytest.fixture(scope="function")
def django_mail_dnsname():
    return "fake-tests.example.com"


@pytest.fixture(autouse=True, scope="function")
def _django_set_urlconf(request):
    """Apply the @pytest.mark.urls marker, internal to pytest-django."""
    marker = request.node.get_closest_marker("urls")
    if marker:
        skip_if_no_django()
        import django.conf

        try:
            from django.urls import clear_url_caches, set_urlconf
        except ImportError:
            # Removed in Django 2.0
            from django.core.urlresolvers import clear_url_caches, set_urlconf

        urls = validate_urls(marker)
        original_urlconf = django.conf.settings.ROOT_URLCONF
        django.conf.settings.ROOT_URLCONF = urls
        clear_url_caches()
        set_urlconf(None)

        def restore():
            django.conf.settings.ROOT_URLCONF = original_urlconf
            # Copy the pattern from
            # https://github.com/django/django/blob/master/django/test/signals.py#L152
            clear_url_caches()
            set_urlconf(None)

        request.addfinalizer(restore)


@pytest.fixture(autouse=True, scope="session")
def _fail_for_invalid_template_variable():
    """Fixture that fails for invalid variables in templates.

    This fixture will fail each test that uses django template rendering
    should a template contain an invalid template variable.
    The fail message will include the name of the invalid variable and
    in most cases the template name.

    It does not raise an exception, but fails, as the stack trace doesn't
    offer any helpful information to debug.
    This behavior can be switched off using the marker:
    ``pytest.mark.ignore_template_errors``
    """

    class InvalidVarException(object):
        """Custom handler for invalid strings in templates."""

        def __init__(self):
            self.fail = True

        def __contains__(self, key):
            """There is a test for '%s' in TEMPLATE_STRING_IF_INVALID."""
            return key == "%s"

        @staticmethod
        def _get_origin():
            stack = inspect.stack()

            # Try to use topmost `self.origin` first (Django 1.9+, and with
            # TEMPLATE_DEBUG)..
            for f in stack[2:]:
                func = f[3]
                if func == "render":
                    frame = f[0]
                    try:
                        origin = frame.f_locals["self"].origin
                    except (AttributeError, KeyError):
                        continue
                    if origin is not None:
                        return origin

            from django.template import Template

            # finding the ``render`` needle in the stack
            frame = reduce(
                lambda x, y: y[3] == "render" and "base.py" in y[1] and y or x, stack
            )
            # assert 0, stack
            frame = frame[0]
            # finding only the frame locals in all frame members
            f_locals = reduce(
                lambda x, y: y[0] == "f_locals" and y or x, inspect.getmembers(frame)
            )[1]
            # ``django.template.base.Template``
            template = f_locals["self"]
            if isinstance(template, Template):
                return template.name

        def __mod__(self, var):
            """Handle TEMPLATE_STRING_IF_INVALID % var."""
            origin = self._get_origin()
            if origin:
                msg = "Undefined template variable '%s' in '%s'" % (var, origin)
            else:
                msg = "Undefined template variable '%s'" % var
            if self.fail:
                pytest.fail(msg)
            else:
                return msg

    if (
        os.environ.get(INVALID_TEMPLATE_VARS_ENV, "false") == "true"
        and django_settings_is_configured()
    ):
        from django.conf import settings as dj_settings

        if dj_settings.TEMPLATES:
            dj_settings.TEMPLATES[0]["OPTIONS"][
                "string_if_invalid"
            ] = InvalidVarException()
        else:
            dj_settings.TEMPLATE_STRING_IF_INVALID = InvalidVarException()


@pytest.fixture(autouse=True)
def _template_string_if_invalid_marker(request):
    """Apply the @pytest.mark.ignore_template_errors marker,
     internal to pytest-django."""
    marker = request.keywords.get("ignore_template_errors", None)
    if os.environ.get(INVALID_TEMPLATE_VARS_ENV, "false") == "true":
        if marker and django_settings_is_configured():
            from django.conf import settings as dj_settings

            if dj_settings.TEMPLATES:
                dj_settings.TEMPLATES[0]["OPTIONS"]["string_if_invalid"].fail = False
            else:
                dj_settings.TEMPLATE_STRING_IF_INVALID.fail = False


@pytest.fixture(autouse=True, scope="function")
def _django_clear_site_cache():
    """Clears ``django.contrib.sites.models.SITE_CACHE`` to avoid
    unexpected behavior with cached site objects.
    """

    if django_settings_is_configured():
        from django.conf import settings as dj_settings

        if "django.contrib.sites" in dj_settings.INSTALLED_APPS:
            from django.contrib.sites.models import Site

            Site.objects.clear_cache()


# ############### Helper Functions ################


class _DatabaseBlockerContextManager(object):
    def __init__(self, db_blocker):
        self._db_blocker = db_blocker

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        self._db_blocker.restore()


class _DatabaseBlocker(object):
    """Manager for django.db.backends.base.base.BaseDatabaseWrapper.

    This is the object returned by django_db_blocker.
    """

    def __init__(self):
        self._history = []
        self._real_ensure_connection = None

    @property
    def _dj_db_wrapper(self):
        from django.db.backends.base.base import BaseDatabaseWrapper

        # The first time the _dj_db_wrapper is accessed, we will save a
        # reference to the real implementation.
        if self._real_ensure_connection is None:
            self._real_ensure_connection = BaseDatabaseWrapper.ensure_connection

        return BaseDatabaseWrapper

    def _save_active_wrapper(self):
        return self._history.append(self._dj_db_wrapper.ensure_connection)

    def _blocking_wrapper(*args, **kwargs):
        __tracebackhide__ = True
        __tracebackhide__  # Silence pyflakes
        raise RuntimeError(
            "Database access not allowed, "
            'use the "django_db" mark, or the '
            '"db" or "transactional_db" fixtures to enable it.'
        )

    def unblock(self):
        """Enable access to the Django database."""
        self._save_active_wrapper()
        self._dj_db_wrapper.ensure_connection = self._real_ensure_connection
        return _DatabaseBlockerContextManager(self)

    def block(self):
        """Disable access to the Django database."""
        self._save_active_wrapper()
        self._dj_db_wrapper.ensure_connection = self._blocking_wrapper
        return _DatabaseBlockerContextManager(self)

    def restore(self):
        self._dj_db_wrapper.ensure_connection = self._history.pop()


_blocking_manager = _DatabaseBlocker()


def validate_django_db(marker):
    """Validate the django_db marker.

    It checks the signature and creates the ``transaction`` and
    ``reset_sequences`` attributes on the marker which will have the
    correct values.

    A sequence reset is only allowed when combined with a transaction.
    """

    def apifun(transaction=False, reset_sequences=False):
        return transaction, reset_sequences

    return apifun(*marker.args, **marker.kwargs)


def validate_urls(marker):
    """Validate the urls marker.

    It checks the signature and creates the `urls` attribute on the
    marker which will have the correct value.
    """

    def apifun(urls):
        return urls

    return apifun(*marker.args, **marker.kwargs)



from django.db import models
from cache import cache_manager
from django.db.models.signals import pre_delete, post_save
from django.dispatch import receiver


class FancyModel(models.Model):
    text = models.TextField()


cache_manager.register_model(
    "fancy_model", FancyModel, values=["content"], instance_values=["pk"]
)

# invalidate cache
@receiver(post_save, sender=FancyModel)
@receiver(pre_delete, sender=FancyModel)
def invalidate_fancymodel(sender, **kwargs):
    cg = cache_manager.get_group("fancy_model")
    cg.invalidate_cache(FancyModel, instance=sender)



__all__ = (
    'WebpackError',
    'WebpackLoaderBadStatsError',
    'WebpackLoaderTimeoutError',
    'WebpackBundleLookupError'
)


class BaseWebpackLoaderException(Exception):
    """
    Base exception for django-webpack-loader.
    """


class WebpackError(BaseWebpackLoaderException):
    """
    General webpack loader error.
    """


class WebpackLoaderBadStatsError(BaseWebpackLoaderException):
    """
    The stats file does not contain valid data.
    """


class WebpackLoaderTimeoutError(BaseWebpackLoaderException):
    """
    The bundle took too long to compile.
    """


class WebpackBundleLookupError(BaseWebpackLoaderException):
    """
    The bundle name was invalid.
    """



from django.forms.widgets import ClearableFileInput
from django.template.loader import render_to_string
from django.utils.html import escape
from django.utils.safestring import mark_safe

try:
    from ..cms.internal_tags.fields import TaggedRelationWidget
    from ..cms.widgets import APIChoiceWidget
except ValueError:
    from cms.internal_tags.fields import TaggedRelationWidget
    from cms.widgets import APIChoiceWidget


class AssetsFileWidget(TaggedRelationWidget):
    crop_link = "crops/{0}/edit/"

    def get_qs(self):
        qs = super().get_qs()
        if self.asset_type:
            qs["ftype"] = self.asset_type
        return qs

    def get_add_qs(self):
        qs = self.get_qs()
        if "ftype" in qs:
            qs["type"] = qs.pop("ftype")
        return qs

    def get_crop_sizes(self):
        from . import get_image_cropper

        sizes = []
        if self.sizes:
            for x in self.sizes:
                crop = get_image_cropper().get_crop_config(x)
                if crop and crop.editable:
                    sizes.append(
                        {
                            "name": crop.name,
                            "width": crop.width,
                            "height": crop.height,
                            "post_link": self.crop_link.format(x),
                        }
                    )
        return sizes

    def render(self, name, value, attrs=None, renderer=None):
        obj = self.obj_for_value(value)
        # Go directly to parent of APIChoiceWidget to get input
        hidden_input = super(APIChoiceWidget, self).render(name, value, attrs=attrs, renderer=None)

        context = {
            "hidden_input": hidden_input,
            "object": obj,
            "asset_type": self.asset_type,
            "asset_tags": self.tags,
            "link": self.get_api_link(),
            "add_link": self.get_add_link(),
            "base_api_link": self._api_link,
            "sizes": self.get_crop_sizes(),
            "required_tags": self.required_tags,
        }
        html = render_to_string("assets/asset_widget.html", context)
        return mark_safe(html)


class RawImageWidget(ClearableFileInput):
    template_with_initial = (
        "%(initial_text)s: %(initial)s %(clear_template)s<br />%(input)s"
    )

    def render(self, name, value, attrs=None, renderer=None):
        thumbnail = None
        data = super().render(name, value, attrs, renderer=None)

        if value and hasattr(value, "admin_url"):
            thumbnail = value.admin_url()
        if thumbnail:
            data = mark_safe(
                '<p class="widget-asset-simple"><span class="widget-asset-simple-preview" style="background-image:url({0})"></span>{1}</p>'.format(
                    escape(thumbnail), data
                )
            )

        return data



from django.conf import settings

# Main Assets Directory. This will be a subdirectory within MEDIA_ROOT.
# Set to None to use MEDIA_ROOT directly
DIRECTORY = getattr(settings, "ASSETS_DIR", "assets")

# Which size should be used as CMS thumbnail for images.
CMS_THUMBNAIL_SIZE = getattr(settings, "ASSETS_CMS_THUMBNAIL_SIZE", "80x80")

# EXTRA SETTINGS
# Convert Filename (UUID)
HASH_FILENAME = getattr(settings, "ASSETS_HASH_FILENAME", True)

# Append a qs to assets urls for cache busting
USE_CACHE_BUST = getattr(settings, "ASSETS_USE_CACHE_BUST", True)

ASSET_MODEL = getattr(settings, "ASSET_MODEL", "assets.Asset")

ASSET_TYPES = getattr(settings, "ASSET_TYPES", None)

DEFAULT_IMAGE_SIZES = {
    "admin": {"width": 100, "height": 100, "editable": False, "upscale": True,},
}

IMAGE_SIZES = getattr(settings, "IMAGE_SIZES", DEFAULT_IMAGE_SIZES)

IMAGE_CROPPER = ".crops.cropper"

CELERY = getattr(settings, "ASSET_CELERY", None)
USE_CELERY_DECORATOR = getattr(settings, "ASSET_USE_CELERY_DECORATOR", False)



from __future__ import absolute_import, division, unicode_literals

from . import base

from collections import OrderedDict


def _attr_key(attr):
    """Return an appropriate key for an attribute for sorting

    Attributes have a namespace that can be either ``None`` or a string. We
    can't compare the two because they're different types, so we convert
    ``None`` to an empty string first.

    """
    return (attr[0][0] or ''), attr[0][1]


class Filter(base.Filter):
    """Alphabetizes attributes for elements"""
    def __iter__(self):
        for token in base.Filter.__iter__(self):
            if token["type"] in ("StartTag", "EmptyTag"):
                attrs = OrderedDict()
                for name, value in sorted(token["data"].items(),
                                          key=_attr_key):
                    attrs[name] = value
                token["data"] = attrs
            yield token



# Generated by Django 1.9 on 2016-02-04 00:37
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="AuthorBase",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="version_twomodels.AuthorBase",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BookBase",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Gallery",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("name", models.CharField(max_length=255)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Review",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("last_save", models.DateTimeField(editable=False)),
                ("text", models.CharField(max_length=255)),
                (
                    "book",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="version_twomodels.Book",
                    ),
                ),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="Store",
            fields=[
                ("last_save", models.DateTimeField(editable=False)),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("published", "published"),
                            ("scheduled", "scheduled"),
                            ("draft", "draft"),
                            ("archived", "archived"),
                        ],
                        editable=False,
                        max_length=50,
                    ),
                ),
                ("last_scheduled", models.DateTimeField(editable=False, null=True)),
                ("date_published", models.DateTimeField(editable=False, null=True)),
                (
                    "user_published",
                    models.CharField(editable=False, max_length=255, null=True),
                ),
                ("vid", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=255)),
                ("books", models.ManyToManyField(to="version_twomodels.BookBase")),
            ],
        ),
        migrations.CreateModel(
            name="StoreBase",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("is_published", models.BooleanField(default=False, editable=False)),
                (
                    "created_date",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("v_last_save", models.DateTimeField(editable=False, null=True)),
            ],
            options={"abstract": False,},
        ),
        migrations.AddField(
            model_name="store",
            name="object",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="version",
                to="version_twomodels.StoreBase",
            ),
        ),
        migrations.AddField(
            model_name="book",
            name="galleries",
            field=models.ManyToManyField(to="version_twomodels.Gallery"),
        ),
        migrations.AddField(
            model_name="book",
            name="object",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="version",
                to="version_twomodels.BookBase",
            ),
        ),
        migrations.AddField(
            model_name="author",
            name="associates",
            field=models.ManyToManyField(blank=True, to="version_twomodels.AuthorBase"),
        ),
        migrations.AddField(
            model_name="author",
            name="object",
            field=models.ForeignKey(
                editable=False,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="version",
                to="version_twomodels.AuthorBase",
            ),
        ),
    ]



from typing import Any
from typing import Generic
from typing import TypeVar

import attr

from _pytest.compat import final
from _pytest.compat import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Type  # noqa: F401 (used in type string)


class PytestWarning(UserWarning):
    """Base class for all warnings emitted by pytest."""

    __module__ = "pytest"


@final
class PytestAssertRewriteWarning(PytestWarning):
    """Warning emitted by the pytest assert rewrite module."""

    __module__ = "pytest"


@final
class PytestCacheWarning(PytestWarning):
    """Warning emitted by the cache plugin in various situations."""

    __module__ = "pytest"


@final
class PytestConfigWarning(PytestWarning):
    """Warning emitted for configuration issues."""

    __module__ = "pytest"


@final
class PytestCollectionWarning(PytestWarning):
    """Warning emitted when pytest is not able to collect a file or symbol in a module."""

    __module__ = "pytest"


@final
class PytestDeprecationWarning(PytestWarning, DeprecationWarning):
    """Warning class for features that will be removed in a future version."""

    __module__ = "pytest"


@final
class PytestExperimentalApiWarning(PytestWarning, FutureWarning):
    """Warning category used to denote experiments in pytest.

    Use sparingly as the API might change or even be removed completely in a
    future version.
    """

    __module__ = "pytest"

    @classmethod
    def simple(cls, apiname: str) -> "PytestExperimentalApiWarning":
        return cls(
            "{apiname} is an experimental api that may change over time".format(
                apiname=apiname
            )
        )


@final
class PytestUnhandledCoroutineWarning(PytestWarning):
    """Warning emitted for an unhandled coroutine.

    A coroutine was encountered when collecting test functions, but was not
    handled by any async-aware plugin.
    Coroutine test functions are not natively supported.
    """

    __module__ = "pytest"


@final
class PytestUnknownMarkWarning(PytestWarning):
    """Warning emitted on use of unknown markers.

    See :ref:`mark` for details.
    """

    __module__ = "pytest"


_W = TypeVar("_W", bound=PytestWarning)


@final
@attr.s
class UnformattedWarning(Generic[_W]):
    """A warning meant to be formatted during runtime.

    This is used to hold warnings that need to format their message at runtime,
    as opposed to a direct message.
    """

    category = attr.ib(type="Type[_W]")
    template = attr.ib(type=str)

    def format(self, **kwargs: Any) -> _W:
        """Return an instance of the warning category, formatted with given kwargs."""
        return self.category(self.template.format(**kwargs))


PYTESTER_COPY_EXAMPLE = PytestExperimentalApiWarning.simple("testdir.copy_example")



# This file is distributed under the same license as the Django package.
#
# The *_FORMAT strings use the Django date format syntax,
# see https://docs.djangoproject.com/en/dev/ref/templates/builtins/#date
DATE_FORMAT = 'j F, Y'
TIME_FORMAT = 'g:i A'
# DATETIME_FORMAT =
YEAR_MONTH_FORMAT = 'F Y'
MONTH_DAY_FORMAT = 'j F'
SHORT_DATE_FORMAT = 'j M, Y'
# SHORT_DATETIME_FORMAT =
FIRST_DAY_OF_WEEK = 6  # Saturday

# The *_INPUT_FORMATS strings use the Python strftime format syntax,
# see https://docs.python.org/library/datetime.html#strftime-strptime-behavior
DATE_INPUT_FORMATS = [
    '%d/%m/%Y',  # 25/10/2016
    '%d/%m/%y',  # 25/10/16
    '%d-%m-%Y',  # 25-10-2016
    '%d-%m-%y',  # 25-10-16
]
TIME_INPUT_FORMATS = [
    '%H:%M:%S',  # 14:30:59
    '%H:%M',  # 14:30
]
DATETIME_INPUT_FORMATS = [
    '%d/%m/%Y %H:%M:%S',  # 25/10/2006 14:30:59
    '%d/%m/%Y %H:%M',  # 25/10/2006 14:30
]
DECIMAL_SEPARATOR = '.'
THOUSAND_SEPARATOR = ','
# NUMBER_GROUPING =



import re
import string

from string import ascii_uppercase

from .. import BaseProvider

localized = True
default_locale = 'en_GB'


class Provider(BaseProvider):
    """
    Provider for IBAN/BBAN: it generates valid (valid length, valid checksum)
    IBAN/BBANs for the given country. But the ids of the banks are random and
    not valid banks! Same for account numbers.
    """

    ALPHA = {c: str(ord(c) % 55) for c in string.ascii_uppercase}

    # see https://en.wikipedia.org/wiki/International_Bank_Account_Number
    bban_format = '????#############'
    country_code = 'GB'

    def bank_country(self):
        return self.country_code

    def bban(self):
        temp = re.sub(r'\?',
                      lambda x: self.random_element(ascii_uppercase),
                      self.bban_format)
        return self.numerify(temp)

    def iban(self):
        bban = self.bban()

        check = bban + self.country_code + '00'
        check = int(''.join(self.ALPHA.get(c, c) for c in check))
        check = 98 - (check % 97)
        check = str(check).zfill(2)

        return self.country_code + check + bban

    def swift8(self, use_dataset=False):
        """Generate an 8-digit SWIFT code

        This method uses :meth:`swift() <faker.providers.bank.Provider.swift>` under the hood
        with the ``length`` argument set to ``8`` and with the ``primary`` argument omitted.
        All 8-digit SWIFT codes already refer to the primary branch/office.

        :sample:
        :sample: use_dataset=True
        """
        return self.swift(length=8, use_dataset=use_dataset)

    def swift11(self, primary=None, use_dataset=False):
        """Generate an 11-digit SWIFT code

        This method uses :meth:`swift() <faker.providers.bank.Provider.swift>` under the hood
        with the ``length`` argument set to ``11``. If ``primary`` is set to ``True``, the SWIFT
        code will always end with ``XXX``. All 11-digit SWIFT codes use this convention to refer
        to the primary branch/office.

        :sample:
        :sample: use_dataset=True
        """
        return self.swift(length=11, primary=primary, use_dataset=use_dataset)

    def swift(self, length=None, primary=None, use_dataset=False):
        """Generate a SWIFT code

        SWIFT codes, reading from left to right, are composed of a 4 alphabet character bank code,
        a 2 alphabet character country code, a 2 alphanumeric location code, and an optional 3
        alphanumeric branch code. This means SWIFT codes can only have 8 or 11 characters, so the
        value of ``length`` can only be ``None`` or the integers ``8`` or ``11``. If the value is
        ``None``, then a value of ``8`` or ``11`` will randomly be assigned.

        Because all 8-digit SWIFT codes already refer to the primary branch/office, the ``primary``
        argument only has an effect if the value of ``length`` is ``11``. If ``primary`` is ``True``
        and ``length`` is ``11``, the 11-digit SWIFT codes generated will always end in ``XXX`` to
        denote that they belong to primary branches/offices.

        For extra authenticity, localized providers may opt to include SWIFT bank codes, location
        codes, and branch codes used in their respective locales. If ``use_dataset`` is ``True``,
        this method will generate SWIFT codes based on those locale-specific codes if included. If
        those codes were not included, then it will behave as if ``use_dataset`` were ``False``,
        and in that mode, all those codes will just be randomly generated as per the specification.

        :sample:
        :sample: length=8
        :sample: length=8, use_dataset=True
        :sample: length=11
        :sample: length=11, primary=True
        :sample: length=11, use_dataset=True
        :sample: length=11, primary=True, use_dataset=True
        """

        if length is None:
            length = self.random_element((8, 11))
        if length not in (8, 11):
            raise AssertionError('length can only be 8 or 11')

        if use_dataset and hasattr(self, 'swift_bank_codes'):
            bank_code = self.random_element(self.swift_bank_codes)
        else:
            bank_code = self.lexify('????', letters=string.ascii_uppercase)

        if use_dataset and hasattr(self, 'swift_location_codes'):
            location_code = self.random_element(self.swift_location_codes)
        else:
            location_code = self.lexify('??', letters=string.ascii_uppercase + string.digits)

        if length == 8:
            return bank_code + self.country_code + location_code

        if primary:
            branch_code = 'XXX'
        elif use_dataset and hasattr(self, 'swift_branch_codes'):
            branch_code = self.random_element(self.swift_branch_codes)
        else:
            branch_code = self.lexify('???', letters=string.ascii_uppercase + string.digits)

        return bank_code + self.country_code + location_code + branch_code



from itertools import product

# Canada uses UPC too
from ..en_US import Provider as BarcodeProvider


class Provider(BarcodeProvider):
    # Source of GS1 country codes: https://gs1.org/standards/id-keys/company-prefix
    local_prefixes = (
        # The source above doesn't specify prefixes 00~01, 06~09 to be used in Canada also,
        # but it's referenced in numerous pages e.g.: https://www.nationwidebarcode.com/upc-country-codes/
        *product((0,), range(2)),
        *product((0,), range(6, 10)),
        (7, 5),
    )
